 /*
 * Program:	VMS mail routines
 *
 * Author:	Yehavi Bourvine, The Hebrew University of Jerusalem.
 *		Internet: Yehavi@VMS.huji.ac.il
 *
 * Modified:	Andy Harper, Kings College London
 *		Most of the original code has been rewritten but I've kept all
 *		acknowledgement notices anyway.
 *
 * Date:	2 August 1994
 * Last Edited:	17 January 1997
 *
 * Copyright 1993 by the University of Washington
 *
 *  Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appears in all copies and that both the
 * above copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the University of Washington not be
 * used in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  This software is made
 * available "as is", and
 * THE UNIVERSITY OF WASHINGTON DISCLAIMS ALL WARRANTIES, EXPRESS OR IMPLIED,
 * WITH REGARD TO THIS SOFTWARE, INCLUDING WITHOUT LIMITATION ALL IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND IN
 * NO EVENT SHALL THE UNIVERSITY OF WASHINGTON BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, TORT
 * (INCLUDING NEGLIGENCE) OR STRICT LIABILITY, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <stdio.h>
#include <types.h>
#include <errno.h>
/* extern int errno;		/* just in case */
#include <signal.h>
#include "mail.h"
#include "osdep.h"
#include <sys/file.h>
#include <sys/stat.h>
/* #include <sys/time.h>		/* must be before osdep.h */
#include "vms_mail.h"
#include "rfc822.h"
#include "misc.h"
#include "dummy.h"
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <descrip.h>
#include <starlet.h>
#include <rms.h>
#include <fdldef.h>

int fdl$create ();
int fdl$generate ();

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

/*
  VMS debugging code
*/
#ifdef LOGTRACE
# define DBGPRINT(level,fmt,a,b,c,d) \
  {char *dbg = getenv("IMAPD_LOGLEVEL"); \
   if (dbg && (atoi(dbg) >= level)) \
     {char tmpmsg[256];sprintf(tmpmsg,fmt,a,b,c,d); mm_dlog(tmpmsg);} \
  }
#else
# define DBGPRINT(level,fmt,a,b,c,d) {}
#endif



/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Miscellaneous bits and pieces
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

					/* Callable MAIL status codes */

#define MAIL$_MSGTEXT   0x007EFD01	/* MAIL - text record returned */
#define MAIL$_NOMOREREC 0x007EDFA8	/* MAIL - No more records */

					/* Subscription database name */
#define MBXDATABASE ".MAILBOXLIST"


					/* Structures */

struct MailBoxList			/* For 'new' mail boxes */
  {
    struct MailBoxList *next;
    char *mailfile;
    char *folder;
  };

struct FolderPattern			/* For folder scanning */
  {
    int count;				/* Incremented if name matches */
    char *pattern;			/* Pattern against which to match folder */
    char *data;				/* Other info we might want */
  };

					/* Abbreviation for item lists */
typedef struct ITEM_LIST item;

					/* Global variables */
static struct MailBoxList *MailBoxes = NULL;
static item   NullItemList[] = {{0, 0, NULL, NULL}};





/* Driver dispatch used by MAIL */

DRIVER vmsmaildriver = {
  "vms_mail",			/* driver name */
  (DRIVER *) NIL,		/* next driver */
  vms_mail_valid,		/* mailbox is valid for us */
  vms_mail_parameters,		/* manipulate parameters */
  vms_mail_find,		/* find mailboxes */
  vms_mail_find_bboards,	/* find bboards */
  vms_mail_find_all,		/* find all mailboxes */
  vms_mail_find_all_bboards,	/* find all bboards */
  vms_mail_subscribe,		/* subscribe to mailbox */
  vms_mail_unsubscribe,		/* unsubscribe from mailbox */
  vms_mail_subscribe_bboard,	/* subscribe to bboard */
  vms_mail_unsubscribe_bboard,	/* unsubscribe from bboard */
  vms_mail_create,		/* create mailbox */
  vms_mail_delete,		/* delete mailbox */
  vms_mail_rename,		/* rename mailbox */
  vms_mail_open,		/* open mailbox */
  vms_mail_close,		/* close mailbox */
  vms_mail_fetchfast,		/* fetch message "fast" attributes */
  vms_mail_fetchflags,		/* fetch message flags */
  vms_mail_fetchstructure,	/* fetch message envelopes */
  vms_mail_fetchheader,		/* fetch message header only */
  vms_mail_fetchtext,		/* fetch message body only */
  vms_mail_fetchbody,		/* fetch message body section */
  vms_mail_setflag,		/* set message flag */
  vms_mail_clearflag,		/* clear message flag */
  vms_mail_search,		/* search for message based on criteria */
  vms_mail_ping,		/* ping mailbox to see if still alive */
  vms_mail_check,		/* check for new messages */
  vms_mail_expunge,		/* expunge deleted messages */
  vms_mail_copy,		/* copy messages to another mailbox */
  vms_mail_move,		/* move messages to another mailbox */
  vms_mail_append,		/* append string message to mailbox */
  vms_mail_gc			/* garbage collect stream */
};

				/* prototype stream */
MAILSTREAM vmsmailproto = {&vmsmaildriver};
                                /* standard driver's prototype */
MAILSTREAM *mailstd_proto = &vmsmailproto;

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Simplify system status checking. Check the status message and print a
	message if it indicates an error. Exit with this status if the abort
        flag is set, otherwise return with LSB of the status.

    Input:
	message		message to display (often name of routine)
	abort		if T then abort, if NIL then just return
	status		status code to display

    Output:
	NONE

    Return Status:
	LSB of status code (1=ok, 0=error)

    Author:
 	Andy Harper, Kings College London, OCT 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int chkstat(char *message, int abort, int status)
  {
     DBGPRINT(23,"%s: exit, status=0x%08x", message, status, 0, 0);
     if ((status & 0x1) == 0)
       {
          char errmess[256];
          sprintf(errmess, "%s: failed with status = 0x%08x", message, status);
          mm_dlog(errmess);

          if (abort)
            { fprintf(stderr,"%s\n", errmess); exit(status); }
       }

     return status & 0x1;
  }

/*
- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
      Case insensitive character match

   Inputs:
	c1		Character to check
	c2		Character to check

   Output:
	NONE

   Return Value:
	T		If a case-insensitive match
	NIL		If no match

   Author:
	Andy Harper, Kings College London, January 1997

- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int chrcasecmp(char c1, char c2)
  {
     return (	(isupper(c1) ? tolower(c1) : c1) ==
		(isupper(c2) ? tolower(c2) : c2)  );
  }


/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Case insensitive string equality test

   Inputs:
	s1		String to check
	s2		String to check

   Output:
	NONE

   Return Value:
	T		If case-insensitive match
	NIL		If no match

   Author:
	Andy Harper, Kings College London, JAN 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int strcasecmp(char *s1, char *s2)
 {
    int n1 = strlen(s1);
    int n2 = strlen(s2);
    if (n1 != n2) return NIL;

    while (n1--)
     if ( !chrcasecmp(*s1++, *s2++) ) return NIL;

    return T;
 }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Match a case-insensitive wildcarded pattern

   Inputs:
	pattern		The pattern to match against
	string		The string to be matched

   Outputs:
	NONE

   Return Value:
	T		If a match
	NIL		If no match

   Comments:
	Primarily required for mailbox location in FIND commands

	Special characters allowed in the pattern are:
		*	Match any string
		? or %	Match exactly one character

        Alphabetic characters are case-insensitive; other characters must match
	themselves exactly.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int matchwild(char *pattern, char *string)
  {
    char *p=pattern, *s=string;
    int i;

    while (*p)
      switch (*p)
        {
         case '*' :				/* Wildcard match */
                    for (i=strlen(s); i>=0; i--)
                      if (matchwild(p+1,s+i)) return T;
                    return NIL;


         case '?':				/* Match any single char */
         case '%':
                   if (!*s++) return NIL; p++; break;


         default:				/* Case-Insensitive match */
                    if ( !chrcasecmp(*p++, *s++) ) return NIL;
        }

    return chrcasecmp(*p, *s);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Create an empty file from the info contained in the FDL string

   Inputs:
	mailfile		Name of mailfile to be created
	fdl			FDL string
	fdlsize			Size of the FDL string

   Outputs:
	NONE

   Return Value:
	T			If file created OK
	NIL			If file not created

   Comments:

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fdl_create_file(char *mailfile, char *fdl, int fdlsize)
  {
     int status, sts, stv;
     int flags   = FDL$M_FDL_STRING;
     struct dsc$descriptor
	fdl_desc = {fdlsize,          DSC$K_DTYPE_T, DSC$K_CLASS_S, fdl},
	filename = {strlen(mailfile), DSC$K_DTYPE_T, DSC$K_CLASS_S, mailfile};

     DBGPRINT(21,"__fdl_create_file: entered, mailfile='%s'",mailfile,0,0,0);
     status = fdl$create(&fdl_desc, &filename, NULL, NULL, NULL, &flags, NULL, NULL, &sts, &stv);
     if (!chkstat("fdl$create", 0, status))
       {
         chkstat("fdl$create sts", 0, sts);
         chkstat("fdl$create stv", 0, stv);
         return NIL;
       }

     DBGPRINT(21,"__fdl_create_file: exit",0,0,0,0);
     return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Obtain an FDL description of a file into a string.

   Inputs:
	file			Name of file
        maxfdlsize		Size of output buffer

   Outputs:
	fdl			Returns the FDL string of the file
	fdlsize			Returns the string length

   Return Value:
	T			Obtained FDL ok
	NIL			FDL unobtainable

   Comments:
	In order to determine the key structure, we first get an XABSUM block
	which contains the number of keys and areas. We then use this to set up
	a dynamic XAB structure for this number of keys and areas.

	We use FDL$GENERATE to get details of the file. Note that the
	documentation on the output parameter is WRONG (as at time of writing)
	as it states that this must be a fixed-length string descriptor. It
	fact, it must be a dynamic string descriptor but with the buffer
	address pointer set to NULL. The buffer is allocated dynamically within
	FDL$GENERATE and must be copied and disposed of manually.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __fdl_generate(char *file, char *fdl, int maxfdlsize, int *fdlsize)
  {
    struct FAB    fab             = cc$rms_fab;
    struct RAB    rab             = cc$rms_rab;
    struct XABSUM xabsum          = cc$rms_xabsum;	/* XAB Summary info */
    struct XAB    *xabptr;

    int noa, nok;
    char keyname[25][32];			/* Space for 25 key names !! */
    struct XABKEY **xabkey;
    struct XABALL **xaball;

    int ret=NIL;				/* Default return status */
    int status;					/* VMS system status returns */


    DBGPRINT(21,"__fdl_generate: entered, file='%s'",file,0,0,0);
						/* Set up filename */
    fab.fab$l_fna = file;
    fab.fab$b_fns = strlen(file);

						/* Open file, get basic info */
    fab.fab$l_xab = &xabsum;
    status = sys$open(&fab);
    if (!chkstat("sys$open", 0, status)) return NIL;

						/* Set up RAB */
    rab.rab$l_fab = &fab;
    status = sys$connect(&rab);
    if (!chkstat("sys$connect", 0, status))
      {status=sys$close(&fab); return NIL;}

						/* Set up the area XABs ... */
    xaball = (struct XABALL **) &fab.fab$l_xab;
    for (noa=0; noa<xabsum.xab$b_noa; noa++)
      {
         *xaball = malloc(sizeof(struct XABALL));
         **xaball= cc$rms_xaball;
         (*xaball)->xab$b_aid = noa;		/* Area number */
         xaball  = (struct XABALL **) &(*xaball)->xab$l_nxt;
      }

						/* Set up the key XABs ... */
    xabkey = (struct XABKEY **) xaball;
    for (nok=0; nok<xabsum.xab$b_nok; nok++)
      {
         *xabkey = malloc(sizeof(struct XABKEY));
         **xabkey= cc$rms_xabkey;
         (*xabkey)->xab$b_ref = nok;		/* Key number */
	 (*xabkey)->xab$l_knm = keyname[nok];	/* Key name */
         xabkey  = (struct XABKEY **) &(*xabkey)->xab$l_nxt;
      }

						/* Re-fill all XAB blocks */
    status = sys$display(&fab);
    if (chkstat("sys$display",0,status))
      {
        struct dsc$descriptor FDLdesc;
        struct FAB *fabptr   = &fab;
        struct RAB *rabptr   = &rab;
        int FDLflags         = FDL$M_FDL_STRING;
						/* Parse FAB/RAB to get FDL */
        FDLdesc.dsc$w_length = maxfdlsize;
        FDLdesc.dsc$b_dtype  = DSC$K_DTYPE_T;
        FDLdesc.dsc$b_class  = DSC$K_CLASS_D;
        FDLdesc.dsc$a_pointer= NULL;
        status = fdl$generate(&FDLflags, &fabptr, &rabptr, NULL, NULL, &FDLdesc, NULL, fdlsize);
        if (chkstat("fdl$generate", 0, status))
          {
						/* Copy to the receive buffer */
            memcpy(fdl, FDLdesc.dsc$a_pointer, *fdlsize);
            fdl[*fdlsize] = '\0';
						/* Dispose of dynamic buffer */
            status = lib$sfree1_dd(&FDLdesc);
            chkstat("lib$sfree1_dd", 0, status);
            ret=T;
          }
      }

						/* Close the file */
    status = sys$close(&fab);
    chkstat("sys$close", 0, status);

						/* Dispose of area/key XAB blocks */
    xabptr = fab.fab$l_xab;
    while (xabptr)
      {
         struct XAB *ptr = xabptr;
         xabptr = xabptr->xab$l_nxt;
         free(ptr);
      }
        
    DBGPRINT(21,"__fdl_generate: exit, status='%d'",ret,0,0,0);
    return ret;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Set message flags in the elt and the cache

   Inputs:
	stream		The mail stream for the mailbox
	i		The message number
	elt		Pointer to the elt for the message
	f		New flags to set

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __internal_setflag(MAILSTREAM *stream, long i, MESSAGECACHE *elt, short f)
  {
      FILECACHE *fc = &LOCAL->msgs[i-1];
      DBGPRINT(21,"__internal_setflag: entered, msgno='%ld', set = '0x%08x'", i, f, 0, 0); 
				/* set all requested flags */
      if (f&fSEEN)     {elt->seen     = T; fc->flags |= fSEEN;}
      if (f&fDELETED)  {elt->deleted  = T; fc->flags |= fDELETED;}
      if (f&fFLAGGED)  {elt->flagged  = T; fc->flags |= fFLAGGED;}
      if (f&fANSWERED) {elt->answered = T; fc->flags |= fANSWERED;}
      DBGPRINT(21,"__internal_setflag: exit", 0, 0, 0, 0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Clear message flags in the elt and the cache

   Inputs:
	stream		The mail stream for the mailbox
	i		The message number
	elt		Pointer to the elt for the message
	f		New flags to clear

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __internal_clearflag(MAILSTREAM *stream, long i, MESSAGECACHE *elt, short f)
  {
      FILECACHE *fc = &LOCAL->msgs[i-1];
      DBGPRINT(21,"__internal_clearflag: entered, msgno='%ld', clear = '0x%08x'", i, f, 0, 0); 
				/* clear all requested flags */
      if (f&fSEEN)     {elt->seen     = NIL; fc->flags &= ~fSEEN;}
      if (f&fDELETED)  {elt->deleted  = NIL; fc->flags &= ~fDELETED;}
      if (f&fFLAGGED)  {elt->flagged  = NIL; fc->flags &= ~fFLAGGED;}
      if (f&fANSWERED) {elt->answered = NIL; fc->flags &= ~fANSWERED;}
      DBGPRINT(21,"__internal_clearflag: exit", 0, 0, 0, 0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	Get the default MAIL directory for the current user

    Input:
	NONE

    Output:
	Mail		Pointer to mail directory buffer pointer; if NULL
			nothing returned here.

    Return Value:
	Mail directory buffer pointer

    Comments:
	Remember what we found on the first successfull call and don't call
	the system routines again if called again.

    Author:
	Andy Harper, Kings College London, OCT 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

char *vms_get_mail_directory(char **Mail)
{
  static char MailDirectory[MAILFILESIZE];	/* Must be static as we return it */
  static int PrevCall = 0;


  DBGPRINT(21,"vms_get_mail_directory: entered",0,0,0,0);

  if (!PrevCall)
    {
      char username[80], host[256], password[80];
      int UserContext=0, length=0, status;

      item InItemList[] = {
	{ 0,              MAIL$_USER_USERNAME,       username,      NULL},
	{ 0, 0, NULL, NULL }};

      item OutItemList[] = {
	{ MAILFILESIZE-1, MAIL$_USER_FULL_DIRECTORY, MailDirectory, &length},
	{ 0, 0, NULL, NULL }};

					/* Ensure we get the right user */
      mm_login(host, username, password, -1L);
      InItemList[0].length = strlen(username);

					/* Quiz MAIL for the directory */
      status = mail$user_begin(&UserContext, NullItemList, NullItemList);
      if (!chkstat("mail$user_begin", 0, status) ) return NULL;
      status = mail$user_get_info(&UserContext, InItemList, OutItemList);
      if (!chkstat("mail$user_get_info", 0, status) ) return NULL;
      status = mail$user_end(&UserContext, NullItemList, NullItemList);
      if (!chkstat("mail$user_end", 0, status) ) return NULL;

					/* Terminate string, store pointer */
      MailDirectory[length] = '\0';
      PrevCall = 1;
    }
  if (Mail) *Mail = MailDirectory;

  DBGPRINT(21,"vms_get_mail_directory: exited, directory='%s'",MailDirectory,0,0,0);
  return MailDirectory;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	Get the login directory (SYS$LOGIN) for the current user

   Input:
	NONE

   Output:
	hp		Pointer to home directory buffer pointer; if NULL
			nothing is returned here.

   Return Value:
	home directory buffer pointer

   Comments:
	Remember what we found on the first call and don't call the system
	routines again if called again.

   Author:
	Andy Harper, Kings College London, 4-OCT-1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *__get_home_directory(char **hp)
  {
     static char home[100], device[32], directory[64];
     static int PrevCall = 0;

     DBGPRINT(21,"__get_home_directory: entered",0,0,0,0);

     if (!PrevCall)
       {
         char host[256], username[256], password[256];
         int status;

         struct DESC usernameD;
         item UserInfo[] = {
	    {sizeof(device),       UAI$_DEFDEV,  device,     0},
	    {sizeof(directory),    UAI$_DEFDIR,  directory,  0},
            {0,                    0,            0,          0}};

					/* Ensure we get the right user */
         mm_login(host, username, password, -1L);
         usernameD.address = ucase( username);
         usernameD.length  = strlen(username);
         usernameD.type    = 0;

					/* Get directory from UAF */
         status = sys$getuai(0,0,&usernameD, UserInfo,0,0,0);
         if (!chkstat("sys$getuai", 0, status) ) return NULL;

					/* Convert to null terminated string */
         strncpy(home, device+1,    (int)device[0]);
         strncat(home, directory+1, (int)directory[0]);
         home [ ((int)device[0]) + ((int)directory[0])] = '\0';
         PrevCall = 1;
       }
     if (hp) *hp = home;

     DBGPRINT(21,"__get_home_directory: exit, directory='%s'",home,0,0,0);
     return home;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Purpose:
	To parse the supplied mailbox specification, supply defaults where
	necessary (mail directory etc.), and split into a mail file and
	folder specification.

        Mailbox is first uppercased (except where quoted).

   Input Formats:
        INBOX			--> [default mail file], NEWMAIL
	folder			--> [default mail file], folder

      The following for unix: compatibility:
	~/folder		--> [relative mail file], folder
        ~/subdir/folder		--> [relative mail file], folder

        The filename part is always "MAIL.MAI" and this should not be
	specified in the pathname.

        A NULL folder name is an error, any other format is invalid


   Input:
	name:		the mailbox name in standard format

   Outputs:
	mailfile	Pointer to VMS filename buffer, or NULL
	folder		Pointer to folder in mailfile buffer, or NULL
	inbox		Pointer to 'is an INBOX' flag

   Returns:
  	NIL mailbox parse error
	T   all ok

   Comments:
	The name of the 'new' mail folder defaults to "NEWMAIL" but can be
	overridden by the logical name IMAPD_INBOXFOLDERNAME.

   Author:
	Andy Harper, Kings College London, 30-SEP-1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/


static int __parse_mailbox_name (char *name, char *mailfile, char *folder, int *inbox)
  {
    static char *DefaultInboxName = "NEWMAIL";		/* Standard 'new' mail folder */
    char mailbox[256], tmpmailfile[MAILFILESIZE];
    char *InboxName, *p, *mbox;
    char c;

    if (!(InboxName=getenv("IMAPD_INBOXFOLDERNAME"))) InboxName=DefaultInboxName;


    DBGPRINT(21,"__parse_mailbox_name: entered, name='%s'", name,0,0,0);

/* -------------------------------------------------------------------------- */
						/* Uppercase, except where quoted */
    p    = name;
    mbox = mailbox;

    while ( (c = *p++) )
      if (c == '"')
        {
          while ( (c = *p++) && (c != '"') )  *mbox++ = c;
          if ( c == '\0') break;
        }
      else
        *mbox++ =  islower(c) ? toupper(c) : c;

    *mbox = '\0';


    mbox = mailbox;

/* -------------------------------------------------------------------------- */
    if ( strcasecmp(mbox, "INBOX") )		/* Special INBOX name */
      {
         strcpy(tmpmailfile, vms_get_mail_directory(NULL) );
         mbox = InboxName;
      }

/* -------------------------------------------------------------------------- */
    else if ( *mbox == '~' )			/* Home-relative specification */
      {
         if (mbox[1] != '/')			/* ... must start '~/'  ... */
           {
             DBGPRINT(21,"__parse_mailbox_name: exit, status=0 : '/' must follow '~'",0,0,0,0);
             return NIL;
           }
         mbox = mbox+2;

         strcpy(tmpmailfile,__get_home_directory(NULL));
         tmpmailfile[strlen(tmpmailfile)-1] = '\0';	/* Strip ']' from end */

						/* strip paths, convert to VMS subdirs */
         while ( (p=strchr(mbox, '/') ) != NULL)
           {
              *p = '\0';			/* Mask out '/'  */
              strcat(tmpmailfile,".");		/* Add subdir separator */
              strcat(tmpmailfile,mbox);		/* Add subdir */
              mbox = p+1;			/* start of next part */
              if (*mbox == '/')			/* Adjacent slashes = error */
                {
                  DBGPRINT(21,"__parse_mailbox_name: exit, status=0 : '//' invalid",0,0,0,0);
                  return NIL;
                }
           }
         strcat(tmpmailfile,"]");			/* Terminate directory part ..*/
      }

/* -------------------------------------------------------------------------- */
    else					/* Simple folder name */
     {
        if (strchr(mbox,'/') != NULL)		/* ... no paths allowed */
         {
           DBGPRINT(21,"__parse_mailbox_name: exit, status=0, absolute path not permitted",0,0,0,0);
           return NIL;
         }
        strcpy(tmpmailfile,vms_get_mail_directory(NULL));
     }


/* -------------------------------------------------------------------------- */
    if (*mbox == '\0')				/* missing folder name */
      {
        DBGPRINT(21,"__parse_mailbox_name: exit, status=0 : missing folder name",0,0,0,0);
        return NIL;
      }


						/* Are we the INBOX ? */
    *inbox = strcasecmp(mbox,       InboxName) && 
             strcasecmp(tmpmailfile,vms_get_mail_directory(NULL));

						/* Append filename */
    strcat(tmpmailfile,"MAIL.MAI");

						/* Copy, if requested */
    if (mailfile) strcpy(mailfile, tmpmailfile);
    if (folder)   strcpy(folder,   mbox);

    DBGPRINT(21,"__parse_mailbox_name: exit, mailfile='%s', folder='%s', inbox='%d'",tmpmailfile, mbox, *inbox, 0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Update the new mail count by subtracting, from the current count, the
	number of messages seen

   Input:
	MessagesSeen		Number of messages seen

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __update_newmail_count(int MessagesSeen)
{
  static char username[256];
  static short Count;
  char password[256],host[256];
  int status, UserContext=0;

  item OldMessageUser[] = {
	{ 0,             MAIL$_USER_USERNAME,         &username, NULL},
	{ 0, 0, NULL, NULL }};

  item OldMessageCount[] = {
	{ sizeof(Count), MAIL$_USER_NEW_MESSAGES,     &Count,    NULL },
	{ 0, 0, NULL, NULL }};

  item NewMessageCount[] = {
	{ 0,             MAIL$_USER_USERNAME,         &username, NULL},
	{ sizeof(Count), MAIL$_USER_SET_NEW_MESSAGES, &Count,    NULL },
	{ 0, 0, NULL, NULL }};

  DBGPRINT(21, "__update_newmail_count: entered, Messages seen = '%d'", MessagesSeen, 0, 0, 0);

		/* Make sure we update the count for the current username */  
  mm_login(host, username, password, -1L);
  OldMessageUser[0].length = strlen(username);
  NewMessageCount[0].length = strlen(username);

		/* Get current count and update by specified amount */
  status = mail$user_begin(&UserContext, NullItemList, NullItemList);
  if (!chkstat("mail$user_begin", 0, status) ) return;

  status = mail$user_get_info(&UserContext, OldMessageUser, OldMessageCount);
  if (!chkstat("mail$user_get_info", 0, status) ) return;
  DBGPRINT(25,"__update_newmail_count: old count = '%d'", Count, 0, 0, 0); 
  Count -= (short) MessagesSeen;
  DBGPRINT(25,"__update_newmail_count: new count = '%d'", Count, 0, 0, 0);
  status = mail$user_set_info(&UserContext, NewMessageCount, NullItemList);
  if (!chkstat("mail$user_set_info", 0, status) ) return;

  status = mail$user_end(&UserContext, NullItemList, NullItemList);
  if (!chkstat("mail$user_end", 0, status) ) return;

  DBGPRINT(21, "__update_newmail_count: exit", 0, 0, 0, 0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Update the message status flags

   Inputs:
	Stream		mail stream
	msgno		Message to update
	syncflag	T   = synchronize external folder
			NIL = just update local flags.

   Output:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

static void __update_message_status (MAILSTREAM *stream, long msgno, long syncflag)
  {
     MESSAGECACHE *elt = mail_elt(stream,msgno);
     FILECACHE    *fc = &LOCAL->msgs[msgno-1];
     short sysflags = 0;
     int status;

     item MessageId[] = {
	{sizeof(msgno),    MAIL$_MESSAGE_ID,    &msgno,    NULL},
	{0,                0,                   NULL,      NULL}};

     item GetFlagsItem[] = {
	{sizeof(sysflags), MAIL$_MESSAGE_RETURN_FLAGS, &sysflags, NULL},
	{0,                0,                   NULL,      NULL}};

     item SetFlagsItem[] = {
	{sizeof(msgno),    MAIL$_MESSAGE_ID,    &msgno,    NULL},
	{sizeof(sysflags), MAIL$_MESSAGE_FLAGS, &sysflags, NULL},
	{0,                0,                   NULL,      NULL}};

     DBGPRINT(21,"__update_message_status: entered, msgno='%d', syncflag='%d'", msgno, syncflag, 0, 0);
     DBGPRINT(25,"__update_message_status: seen='%d', deleted='%d', answered='%d',flagged='%d'", elt->seen, elt->deleted, elt->answered, elt->flagged);

					/* Update cache flags from elt */
     DBGPRINT(25,"__update_message_status: old cache flags = '0x%08x'",fc->flags, 0, 0, 0);

					/* If message now seen but wasn't before, update new mail count */
     if (elt->seen)
       { if ( syncflag && !((fc->flags)&fSEEN) ) __update_newmail_count(1);
         __internal_setflag(stream, msgno, elt, fSEEN);
       }
     else
       __internal_clearflag(stream, msgno, elt, fSEEN);

     if (elt->deleted)
       __internal_setflag(stream, msgno, elt, fDELETED);
     else
       __internal_clearflag(stream, msgno, elt, fDELETED);

     if (elt->flagged)
       __internal_setflag(stream, msgno, elt, fFLAGGED);
     else
       __internal_clearflag(stream, msgno, elt, fFLAGGED);

     if (elt->answered)
       __internal_setflag(stream, msgno, elt, fANSWERED);
     else
       __internal_clearflag(stream, msgno, elt, fANSWERED);

     DBGPRINT(25,"__update_message_status: new cache flags = '0x%08x'", fc->flags, 0, 0, 0);


					/* Update mailfile with message flags */
     if (syncflag)
       {
         DBGPRINT(25,"__update_message_status: syncing message %d flags with mailfile", msgno, 0, 0, 0);
         status = mail$message_get(&LOCAL->msg_context, MessageId, GetFlagsItem);
         if (!chkstat("mail$message_get", 0, status)) return;
         DBGPRINT(25,"__update_message_status: current sys flags = '0x%08x' (seen='%d', flagged='%d', answered='%d')", sysflags, elt->seen, elt->flagged, elt->answered);

         if (elt->seen)    sysflags &= ~MAIL$M_NEWMSG;   else sysflags |= MAIL$M_NEWMSG;
         DBGPRINT(25,"__update_message_status: seen='%d', new sysflags='0x%08x'", elt->seen, sysflags, 0, 0);
         if (elt->flagged) sysflags |= MAIL$M_MARKED;   else sysflags &= ~MAIL$M_MARKED;
         DBGPRINT(25,"__update_message_status: flag='%d', new sysflags='0x%08x'", elt->flagged, sysflags, 0, 0);
         if (elt->answered)sysflags |= MAIL$M_REPLIED;  else sysflags &= ~MAIL$M_REPLIED;
         DBGPRINT(25,"__update_message_status: answ='%d', new sysflags='0x%08x'", elt->answered, sysflags, 0, 0);

         DBGPRINT(25,"__update_message_status: new     sys flags = '0x%08x'", sysflags, 0, 0, 0);
         status = mail$message_modify(&LOCAL->msg_context, SetFlagsItem, NullItemList);
         if (!chkstat("mail$message_modify", 0, status)) return;
         DBGPRINT(25,"__update_message_status: message/mailfile sync complete",0,0,0,0);
       }

    DBGPRINT(21,"__update_message_status: exit, message '%d' status updated",msgno,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Determine the VMS MAIL transport prefix

   Input:
	NONE

   Output:
	NONE

   Return Value:
	Pointer to prefix name, or NULL

   Comments:
	The prefix is determined from two logical names:
	  *  PINE_MAIL_PROTOCOL
          *  MAIL$INTERNET_TRANSPORT

	The latter is usually defined for OpenVMS 6.2 and above

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *__transport_name(void)
  {
     char *p;

     if      ((p=getenv("PINE_MAIL_PROTOCOL"))      != NULL) return p;
     else if ((p=getenv("MAIL$INTERNET_TRANSPORT")) != NULL) return p;

     return NULL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

  Purpose:
    Help Map DECnet style addresses (namespace:.node::user, node::user and user)
    into RFC822 style, using logical names:

      IMAPD_NODE_namespace_node	If both namespace and node   [ namespace:.node::user ]
      IMAPD_NODE_node		If nodename only             [ node::user ]
      IMAPD_NODE		If no namespace or nodename  [ username]

    What we return is either the translation of the appropriate logical name
    the original nodename, or (as a last resort) the current system name, for
    use in constructing the user@node RFC822 format

  Author:
    Andy Harper, Kings College London, October 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static char *__map_nodename(char *namespace, char *nodename)
  {
     char *p, *ns=namespace, *nn=nodename, logname[256]="IMAPD_NODE";
     int c, s = strlen(logname);


					/* Add namespace part to logical name */
     if (ns && *ns)
       {
         logname[s++] = '_';
         while (c = *ns++)
            logname[s++] = islower(c) ? toupper(c) : c;
       }

					/* Add nodename part to logical name */
     if (nn && *nn)
       {
         logname[s++] = '_';
         while (c = *nn++)
            logname[s++] = islower(c) ? toupper(c) : c;
       }
					/* Finish off the string */
     logname[s]   = '\0';

					/* Return any translation */
     if ((p=getenv(logname)) != NULL)
        return p;

					/* .. or the original nodename */
     if (nodename != NULL)
        return nodename;

					/* .. or the local hostname */
     return mylocalhost();
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Given an address string that resembles one of the standard VMS MAIL
	From: headers, return an RFC822 equivalent.

   Conversions:
	xyz%"..."		-->	...
	abc@def			-->	abc@def
	user			-->	username@node
	node::user		-->	user@node
	namespace:.node::user	-->	user@node

	In the last three cases, the nodename/namespace is optionally mapped
	by logicals into an alternative nodename string

   Author:
	Andy Harper, Kings College London, October 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static char *__to_rfc822(char *decnetaddress, char *rfc822address)
  {
    char *c, *namespace = NULL, *nodename = NULL;

					/* Look for  xyz%"...."     */
    if ((c=strchr(decnetaddress,'%')) != NULL)
      {
        if ( *++c == '"')
           {
             c++;
             if (strrchr(c,'"') != NULL) *strrchr(c,'"') = '\0';
             
           }
        strcpy(rfc822address,c);
        return rfc822address;
      }

					/* Look for internet style addresses */
    if (strchr(decnetaddress,'@') != NULL)
      {
        strcpy(rfc822address,decnetaddress);
        return rfc822address;
      }

					/* Look for DECnet Phase V namespace */
    if ((c=strstr(decnetaddress,":.")) != NULL)
      {
        *c = '\0';
        namespace = decnetaddress;
        decnetaddress = c + 2;
      }

					/* Look for a DECnet nodename */
    if ((c=strstr(decnetaddress,"::")) != NULL)
      {
         *c = '\0';
         nodename = decnetaddress;
         decnetaddress = c + 2;
      }

					/* Convert to RFC822 */
    strcpy(rfc822address,decnetaddress);
    strcat(rfc822address,"@");
    strcat(rfc822address, __map_nodename(namespace, nodename));

    return rfc822address;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	VMS's DECnet style addressing will confuse things, so we'll reformat
	them into a more RFC822-like syntax here.

   Input formats:
        user				(no DECnet)
        node::user			(DECnet IV)
        namespace:.node::user		(DECnet V)

        plus, comma separated list of the above

   Output formats:
	user@nodename

   Comments:
	The address line always end in '\n'

	If any translated address will not fit into the output record, it
	is silently dropped. This prevents writing into random memory!

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __rewrite_address(char *record, int *size, int maxlen)
 {
   char *p, *a, *address, fieldcontent[256], rfc822[256];
   int fcsize;

					/* separate field and content */
   p = strchr(record,':')+1;		/* ... find field terminator */
   fcsize = (*size-1) - (p-record);	/* ... size of field contents */
   strncpy(fieldcontent, p+1, fcsize);	/* ... copy content, leave field */
   fieldcontent[fcsize-1] = '\0';	/* Terminate and strip out \n */

   *p++ = ' ';				/* ... Change tab to space */


					/* Scan list of addresses */
   address = fieldcontent;
   while (*address)
     {
					/* Last username in list ??? */
       if ((a=strchr(address,',')) == NULL)
        {
           __to_rfc822(address, rfc822);
           if ((p-record) + strlen(rfc822) < maxlen-1)
              {strcpy(p, rfc822); p+=strlen(rfc822);}
           break;
        }

					/* Not last !! */
       *a = '\0';
       __to_rfc822(address,rfc822);
       if ((p-record) + strlen(rfc822) + 1 < maxlen-1)
         {strcpy(p, rfc822); p+=strlen(rfc822); *p++ = ',';}
       address = a + 1;			/* Go to next username listed */
     }

    *p++ = '\n';			/* Add trailing newline */
    *size = (p-record);			/* return new size */
 }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Replace the VMS 'Subj:' header by the RFC822 equivalent 'Subject:'

   Inputs:
	header		The header to be modified
	size		Pointer to current length of header
	maxlen		Maximum space available in header buffer

   Output:
	header		Modified in situ as necessary
	size		Modified to reflect new size as necessary

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __rewrite_subject(char *header, int *size, int maxlen)
  {
    static char oldsubj[] = "Subj:\t", newsubj[]="Subject: ";
    static int oldlen=sizeof(oldsubj)-1, newlen=sizeof(newsubj)-1;
    static int diff = sizeof(newsubj) - sizeof(oldsubj);

				/* Replace if enough free space */
    if (*size+diff <= maxlen)
      {
				/* Move field contents up, replace field name */
        memmove(header+newlen, header+oldlen, *size-oldlen);
        memmove(header,newsubj,newlen);

        *size += diff;
      }
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Rewrite standard VMS MAIL headers into a more standard RFC822 form

   Inputs:
	header		The header to be modified
	size		Pointer to current length of header
	maxlen		Maximum space available in header buffer

   Output:
	header		Modified in situ as necessary
	size		Modified to reflect new size as necessary

   Return Value:
	NONE

   Comments:
	A FROM: field has additional info (personal name and date) that we
	don't want, so this must be eliminated.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __rewrite_vms_headers(char *header, int *size, int maxlen)
  {
     if (search(header, *size, "from:", 5) == T)
       {
         int i   = *size;
         char *p = header;
					/* Remove personal name/date fields */
         while (i--)
            if (*p++ == ' ') break;
					/* Terminate properly */
         *p++  = '\n';
         *size = p-header;
         __rewrite_address(header, size, maxlen);
       }
     else if (search(header, *size, "to:", 3) == T)
         __rewrite_address(header, size, maxlen);
     else if (search(header, *size, "cc:", 5) == T)
         __rewrite_address(header, size, maxlen);
     else if (search(header, *size, "subj:", 5) == T)
         __rewrite_subject(header, size, maxlen);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    Purpose:
	This routine examines the header record and returns T if it looks like
	an RFC822 header and NIL otherwise. Address fields must contain an
	'@' symbol to reduce the risk of lines starting with a valid header
        being mistakenly interpreted.

    Inputs:
	record		Header record buffer
	size		Size of the line in the buffer

    Output:
	NONE

    Return Value:
	T		Record appears to be an RFC822 mail header
	NIL		Record is not a header

    Author:
	Andy Harper, Kings College London, Sep 30 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __is_rfc822_header(char *record, int size)
  {
    int ats = search(record, size, "@", 1);

  /* Standard RFC822 headers */
    if (search(record, size, "to:",                       3 ) == T && ats) return T;
    if (search(record, size, "cc:",                       3 ) == T && ats) return T;
    if (search(record, size, "via:",                      4 ) == T       ) return T;
    if (search(record, size, "bcc:",                      4 ) == T && ats) return T;
    if (search(record, size, "date:",                     5 ) == T       ) return T;
    if (search(record, size, "from:",                     5 ) == T && ats) return T;
    if (search(record, size, "sender:",                   7 ) == T && ats) return T;
    if (search(record, size, "subject:",                  8 ) == T       ) return T;
    if (search(record, size, "comment:",                  8 ) == T       ) return T;
    if (search(record, size, "received:",                 9 ) == T       ) return T;
    if (search(record, size, "reply-to:",                 9 ) == T && ats) return T;
    if (search(record, size, "message-id:",               11) == T       ) return T;
    if (search(record, size, "return-path:",              12) == T && ats) return T;
    if (search(record, size, "mime-version:",             13) == T       ) return T;
    if (search(record, size, "content-type:",             13) == T       ) return T;
    if (search(record, size, "read-receipt-to:",          16) == T && ats) return T;

/* The following deal with mail posted via newsgroups */
    if (search(record, size, "path:",                     5 ) == T       ) return T;
    if (search(record, size, "lines:",                    6 ) == T       ) return T;
    if (search(record, size, "control:",                  8 ) == T       ) return T;
    if (search(record, size, "approved:",                 9 ) == T && ats) return T;
    if (search(record, size, "newsgroups:",               11) == T       ) return T;
    if (search(record, size, "followup-to:",              12) == T       ) return T;
    if (search(record, size, "distribution:",             13) == T       ) return T;
    if (search(record, size, "organization:",             13) == T       ) return T;
    if (search(record, size, "relay-version:",            14) == T       ) return T;


  /* Some common extensions */
    if (search(record, size, "x-sender:",                  9) == T && ats) return T;
    if (search(record, size, "x-mailer:",                  9) == T       ) return T;
    if (search(record, size, "x-authentication:",         17) == T       ) return T;
    if (search(record, size, "x-authentication-warning:", 25) == T       ) return T;

    return NIL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Get information about a mail message

   Inputs:
	stream		Mail stream
	msgno		Number of message for which info required

   Outputs:
	size		Pointer to maximum byte size of message
	flags		Pointer to message status flags
	arrivaldate	Pointer to arrival date string pointer

   Return Value:
        NONE

   Comments:
	The 'size' is returned as a record count so we convert it a maximum
	byte count by multiplying by the maximum record count. In practice,
	records will be much less than this byte count but we wont ever go
	bigger so allocating this muchy buffer space later is pretty safe (if a
	little excessive). We can't find the actual byte count without reading
	the whole message.

	The meaning of the message flag bits are as follows (taken from the
	maildef.h include header):

	BIT	DECIMAL	SYMBOLIC NAME		MEANING
	  0	1	MAIL$M_NEWMSG		A NEW message
	  1	2	MAIL$M_REPLIED		Message has been REPLIED to
	  2	4	MAIL$M_DEL		Message has been deleted
	  3	8	MAIL$M_EXTMSG		Message text is in external file
	  4	16	 :
	  5	32	 :  Miscellaneous status, not useful to us
	  6	64	 :
	  7	128	MAIL$M_MARKED		Message has been marked


   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static void __message_info(MAILSTREAM *stream, long msgno, unsigned long *size, short *flags, char **arrivaldate)
  {
    static char DateBuff[256];
    static unsigned int RecCount, DateLen=0;
    int status;

    item MessageIdItem[] = {
	{sizeof(msgno),       MAIL$_MESSAGE_ID,           &msgno,    NULL },
	{0, 0, NULL, NULL }};

    item MessageInfoItem[] = {
	{sizeof(RecCount),    MAIL$_MESSAGE_SIZE,         &RecCount, NULL },
	{sizeof(DateBuff)-1,  MAIL$_MESSAGE_DATE,         DateBuff,  &DateLen},
	{sizeof(*flags),      MAIL$_MESSAGE_RETURN_FLAGS, flags,     NULL },
	{0, 0, NULL, NULL }};

  DBGPRINT(21,"__message_info: entered, mailbox='%s', msgno='%ld'", stream->mailbox, msgno,0,0);


					/* Obtain message info */
  status = mail$message_get(&LOCAL->msg_context, MessageIdItem, MessageInfoItem);
  chkstat("mail$message_get", 1, status);


					/* Convert/copy message info */
  *size = (RecCount+4)*MAXRECORDSIZE;	/* Allow 4 extra for VMS MAIL headers */
  *arrivaldate = DateBuff;
  DateBuff[DateLen] = '\0';

  DBGPRINT(21, "__message_info: exit, records='%u', size='%lu', flags='0x%08x', date='%s'", RecCount, *size, *flags, DateBuff);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Obtain the next record of the current message from the MAIL folder

   Input:
	stream		The mail stream of the folder

   Output:
	record		Pointer to text buffer for the record

   Return Value:
	record length (including newline) or 0 if EOF/error

   Comments:
	The message context should have been previously selected.

	The record is terminated by a newline character

   Author:
	Andy Harper, Kings College London, 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

static int __message_readline(MAILSTREAM *stream, char *record)
  {
    static int RecordLength;
    int i, status;

    static item GetMessageTextItem[] = {
        { sizeof(int),   MAIL$_MESSAGE_CONTINUE, NULL, NULL },
        { 0, 0, NULL, NULL }};

    static item RecordItem[] = {
        { MAXRECORDSIZE, MAIL$_MESSAGE_RECORD,   NULL, &RecordLength },
        { 0, 0, NULL, NULL }};


/*  Get the next line of data from the message */
    RecordItem[0].buffer = record;
    status = mail$message_get(&LOCAL->msg_context, GetMessageTextItem, RecordItem);
    if (status == MAIL$_MSGTEXT)	/* Found record OK  */
      { record[RecordLength]='\n'; return RecordLength+1; }

    if (status == MAIL$_NOMOREREC)	/* End of message */
      { return 0; }

					/* Some unknown error !!! */
    chkstat("mail$message_get",0,status);
    DBGPRINT(25,"__message_readline: unexpected error '0x%08x' reading message",status,0,0,0);
    return 0;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  Purpose:
    The message is read, line by line, from the mail file and, under VMS, has
    this format:
         VMS MAIL headers (From, To, CC, Subj)
         Blank Line
         [Optional RFC822 headers, if delivered by a network transport,
          Blank line]
         Body of text

    This routine reads the message and separates it into headers
    and text parts. Where RFC822 headers are (apparently) present, they
    overwrite the VMS MAIL headers, otherwise the VMS MAIL headers are
    reformatted into an rfc822-like syntax where possible.

   Inputs:
	stream		The mailstream
	msgno		The message to read

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	This routine takes care to avoid the use of the C null string
	terminator where possible in order to allow null bytes to appear in a
	message (this may not be usual but it's a restriction that should not
	be imposed, so we don't).

	We use the LOCAL->buf structure as a temporary buffer in which to read
	the message. It is resized if needed.

        The message is cached on the first access; subsequent reads do nothing

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __read_message (MAILSTREAM *stream, long msgno)
  {
     FILECACHE     *fc = &LOCAL->msgs[msgno-1];
     MESSAGECACHE  *elt;
     char          *mbuf, *mend, *mbptr, *starthdr, *starttext;
     char          *deliverydate;
     unsigned long size;
     int           len;     
     short         flags;


     DBGPRINT(21,"__read_message: entered, mailbox='%s', msgno='%ld'",stream->mailbox, msgno,0,0);

						/* If cached, nothing to do */
     if (fc->text && fc->header)
       {
         DBGPRINT(21,"__read_message: exit, message in cache",0,0,0,0);
         return;
       }


  /* ---------------- Get basic info about message ----------------- */

     elt = mail_elt(stream,msgno);
     __message_info(stream, msgno, &size, &flags, &deliverydate);

     elt->seen     = ((flags & MAIL$M_NEWMSG)  ? NIL : T);
     elt->answered = ((flags & MAIL$M_REPLIED) ? T : NIL);
     elt->flagged  = ((flags & MAIL$M_MARKED)  ? T : NIL);
     __update_message_status(stream,msgno,NIL);


  /* ----------------  Local buffer setup and initialization ------ */
						/* Resize buffer if needed */
     if (size >= LOCAL->buflen)
       {
         fs_give((void **) &LOCAL->buf);
         LOCAL->buf = (char *) fs_get((LOCAL->buflen=size)+1);
       }

						/* Initialize buffer pointers */
     mbptr = mbuf  = LOCAL->buf;
     mend  = mbuf + size + 1;   


  /* -----------------   VMS MAIL headers  ----------------------- */

     DBGPRINT(25,"__read_message: getting VMS MAIL headers",0,0,0,0);
     starthdr = mbptr;				/* Note start of headers */
     len=__message_readline(stream, mbptr);	/* First record */
     while ( len > 1 )				/* Continue until blank line */
       {
          __rewrite_vms_headers(mbptr,&len,mend-mbptr);
          len=__message_readline(stream, mbptr+=len);
       }

						/* Add delivery date field */
     sprintf(mbptr,"Date: %s\n", deliverydate);
     mbptr += strlen(mbptr);
     *mbptr++ = '\n';				/* terminate headers */

						/* set date into elt */
     mail_parse_date(elt, deliverydate);

  /* -----------------   Possible RFC822  headers  ----------------------- */

     DBGPRINT(25,"__read_message: checking for RFC822 headers",0,0,0,0);
     len = __message_readline(stream, mbptr);
     if ( (len > 0) && __is_rfc822_header(mbptr,len) )
       {
         starthdr = mbptr;			/* Note new start of headers */
         while (len > 1)			/* Continue until blank line */
            len = __message_readline(stream, mbptr+=len);
         *mbptr++ = '\n';			/* terminate headers */

						/* First body line */
         len = __message_readline(stream, mbptr);
       }

  /* -----------------   Message body  ----------------------- */

     DBGPRINT(25,"__read_message: reading message body",0,0,0,0);
     starttext = mbptr;				/* Remember where body starts */
     while (len > 0)				/* Continue until EOM */
        len = __message_readline(stream, mbptr+=len);


  /* -----------------  Save in cache ---------------------------------- */

						/* Save headers in cache */
     DBGPRINT(25,"__read_message: caching headers, size='%d'",starttext-starthdr,0,0,0);
     fc->headersize = (starttext-starthdr);	/* How big ?? */
     fc->header     = (char *) fs_get(fc->headersize+1);
     memmove(fc->header, starthdr, fc->headersize);

						/* Save body in cache */
     DBGPRINT(25,"__read_message: caching body, size='%d'",mbptr-starttext,0,0,0);
     fc->size  = (mbptr - starttext);		/* How big ?? */
     fc->text  = (char *) fs_get(fc->size + 1);
     memmove(fc->text, starttext, fc->size);

     DBGPRINT(21,"__read_message, exit, message size='%ld'",fc->headersize+fc->size,0,0,0);
  }

/* VMS/MAIL mail lock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name buffer
 *	    type of locking operation (LOCK_SH or LOCK_EX)
 * Returns: file descriptor of lock or -1 if failure
 */

int vms_mail_lock (FILE *fd, char *lock, int op)
{
 return 0;		/* no locking needed */
}


/* VMS/MAIL mail unlock file for parse/append permission
 * Accepts: file descriptor
 *	    lock file name from vms_mail_lock()
 */

void vms_mail_unlock (FILE *fd, char *lock)
{
  return;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Add a newly created mailfile/folder combination to the list of 'new'
	mailboxes.

   Inputs:
	mailfile		Pointer to name of mailfile
	folder			Pointer to name of folder in mailfile

   Outputs:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __Add_New_Mailbox(char *mailfile, char *folder)
  {
     struct MailBoxList *mb = fs_get( sizeof(struct MailBoxList) );

     DBGPRINT(21,"__Add_New_Mailbox, entered, mailfile='%s',folder='%s'",mailfile,folder,0,0);

					/* Link to head of chain */
     mb->next = MailBoxes;

					/* Copy items into structure */
     mb->mailfile = fs_get( strlen(mailfile) + 1);
     strcpy(mb->mailfile, mailfile);
     mb->folder   = fs_get( strlen(folder)   + 1);
     strcpy(mb->folder, folder);

					/* Link master pointer */
     MailBoxes = mb;

     DBGPRINT(21,"__Add_New_Mailbox: exit",0,0,0,0);
  }



/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Search the list of 'new' mailboxes

   Inputs:
	mailfile		Pointer to name of mail file
	folder			Pointer to name of folder

   Outputs:
	NONE

   Return Value:
	T			if found
	NIL			if not found

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __search_new_mailboxes(char *mailfile, char *folder)
  {
     struct MailBoxList *mb;

     DBGPRINT(21,"__search_new_mailboxes: entered, mailfile='%s', folder='%s'",mailfile,folder,0,0);
     for (mb=MailBoxes; mb; mb=mb->next)
       {
         DBGPRINT(25,"__search_new_mailboxes: checking mailfile='%s', folder='%s'",mb->mailfile,mb->folder,0,0);
         if ( (strcmp(mb->mailfile, mailfile) == 0) &&
              (strcmp(mb->folder,   folder)   == 0) )
           {
             DBGPRINT(21,"__search_new_mailboxes: exit, found",0,0,0,0);
             return T;
           }
        }

     DBGPRINT(21,"__search_new_mailboxes: exit, not found",0,0,0,0);
     return NIL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Initialize the mail stream's local text cache. This cache remembers the
	headers and message body, in text form, for each message as it's read.

   Input:
	stream		The mail stream for which the cache is to be created

   Output:
	NONE

   Return Value:
	NONE

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __init_local_cache(MAILSTREAM *stream)
  {
    long i;
					/* Allocate the message cache */
    LOCAL->msgs = NULL;
    if (stream->nmsgs == 0) return;
    LOCAL->msgs = fs_get(stream->nmsgs * sizeof(FILECACHE));

					/* Clear all entries to null pointers */
    for(i = 0; i < stream->nmsgs; i++) {
	LOCAL->msgs[i].flags      = 0;
        LOCAL->msgs[i].size       = 0;
	LOCAL->msgs[i].text       = NULL;
        LOCAL->msgs[i].headersize = 0;
        LOCAL->msgs[i].header     = NULL;
    }
  }



/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Remove the mail stream's local text cache.

   Input:
	stream		The mail stream for which the cache is to be removed

   Output:
	NONE

   Return Value:
	NONE

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __clear_local_cache(MAILSTREAM *stream)
  {
    long i;
				/* Free all occupied entries */
    for(i = 0; i < stream->nmsgs; i++)
      {
	if(LOCAL->msgs[i].text != NULL)
		fs_give((void **) &LOCAL->msgs[i].text);
	if(LOCAL->msgs[i].header != NULL)
		fs_give((void **) &LOCAL->msgs[i].header);
      }
				/* Free the cache table itself */
    fs_give((void **) &LOCAL->msgs);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Action routine to process each folder found by the FIND ALL command
	It simply notifies the application of the folder name, via the
	mm_mailbox callback.

   Inputs:
	fpi			Folder pattern information
	FolderName		Name of latest folder found in mailbox

   Outputs:
	fpi			As above; contains return status flag

   Return Value:
	Ignored

   Side Effects:
	Calls mm_mailbox with the foldername; the application may do something
	with this name.

   Comments:
	The fpi structure contains the pattern to be matched against and
	returns the count of matches found. Any folder which matches the
	pattern is notified to the application using the mm_mailbox callback.
	The matched name is lowercased and prefixed by the string in the data
	field of the fpi structure

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __match_folders(struct FolderPattern *fpi, struct DESC *FolderName)
  {
    char mailbox[MAILFILESIZE];

    if ( (FolderName==NULL) || (FolderName->length == 0) )/* End of list */
      {
        DBGPRINT(21,"__match_folders: end of folder list",0,0,0,0);
        return 0;
      }

					/* Get the file mailbox name */
    strcpy(mailbox,fpi->data);
    strncat(mailbox, FolderName->address, FolderName->length);
    lcase(mailbox);

					/* Register if it matches the pattern */
    if (matchwild(fpi->pattern, mailbox) == T)
      {
        fpi->count++;
        mm_mailbox( mailbox );
      }

    DBGPRINT(21,"__match_folders: pattern='%s', count='%3d', folder='%.*s'", fpi->pattern, fpi->count, FolderName->length, FolderName->address);
    return 1;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Action routine to locate a particular folder within a mail file.

   Inputs:
	fpi			Folder pattern information
	FolderName		Name of latest folder found in mailbox

   Outputs:
	fpi			As above; contains return status flag

   Return Value:
	Ignored

   Comments:
	The fpi structure contains the folder name required. If found, the
	count field is incremented.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __find_folder(struct FolderPattern *fpi, struct DESC *nextfolder)
  {
				 	/* End of list */
    if( (nextfolder==NULL)  ||  (nextfolder->length == 0) )
      {
        DBGPRINT(21,"__find_folder: end of folder list",0,0,0,0);
        return 0;
      }

					/* See if this matches what we want */
    if (nextfolder->length == strlen(fpi->pattern))
      if ( strncmp(nextfolder->address, fpi->pattern, nextfolder->length) == 0)
        fpi->count++;


    DBGPRINT(21,"__find_folder: count '%3d', folder '%.*s'", fpi->count, nextfolder->length, nextfolder->address, 0);
    return 1;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Scan the folders in a mailfile and call an action routine to deal
	with each one. This is really a simple jacket routine for the
	mail$mailfile_info_file routine.

   Inputs:
	context		Address of the mailfile context
	routine		Address of function to call
	userdata	Address of userdata to pass

   Outputs:
	NONE

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static void __scan_folders(int *context, int (*action)(), void *userdata)
  {

    int status;

    item ScanFolders[] = {
	{sizeof(action),   MAIL$_MAILFILE_FOLDER_ROUTINE, action,    NULL},
	{sizeof(userdata), MAIL$_MAILFILE_USER_DATA,      userdata, NULL},
        {0,                0,                             NULL,      NULL}};

    DBGPRINT(21, "__scan_folders: entered, action = '0x%08x', userdata='0x%08x'",action, userdata, 0, 0);
    status = mail$mailfile_info_file(context, ScanFolders, NullItemList);
    chkstat("mail$mailfile_info_file", 1, status);
    DBGPRINT(21, "__scan_folders: exit",0,0,0,0);
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy a message from the current message/folder to a new mailbox

   Inputs:
	stream		The mail stream for the message source
	msgno		The message number
	outmfile	Ptr to name of destination mailfile
	outfolder	Ptr to name of destination folder

   Outputs:
	NONE

   Return Value:
	T		If copy OK
	NIL		If copy not OK

   Author:
	Andy Harper, Kings College London, November 1996

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int vms_mail_message_copy(MAILSTREAM *stream, long msgno, char *outmfile, char *outfolder)
  {
    static int FileCreated, FoldCreated;
    int status;

    item CopyFolderItem[] = {
	{strlen(outfolder),   MAIL$_MESSAGE_FOLDER,         outfolder,    NULL},	/* must be 1st,foldername */
	{strlen(outmfile),    MAIL$_MESSAGE_FILENAME,       outmfile,     NULL},	/* must be 2nd */
	{sizeof(msgno),       MAIL$_MESSAGE_ID,             &msgno,       NULL},	/* Message number to move */
	{0, 0, NULL, NULL } };
    
    item CreatedItemList[] = {
	{sizeof(FileCreated), MAIL$_MESSAGE_FILE_CREATED,   &FileCreated, NULL},
	{sizeof(FoldCreated), MAIL$_MESSAGE_FOLDER_CREATED, &FoldCreated, NULL},
	{0, 0, NULL, NULL } };

    DBGPRINT(21,"vms_mail_message_copy: entered, source='%s', msgno='%ld', destination='%s -- %s'", stream->mailbox, msgno, outmfile, outfolder);
    
    status = mail$message_copy(&LOCAL->msg_context, CopyFolderItem, CreatedItemList);
    if (!chkstat("mail$message_copy", 0, status) ) return NIL;
    if (FileCreated) mm_log("Created mailbox file", WARN);
    if (FoldCreated) mm_log("Created mailbox folder", WARN);

    DBGPRINT(21,"vms_mail_message_copy: exit",0,0,0,0);
    return T;
  }

/* VMS/MAIL copy messages
 * Accepts: MAIL stream
 *	    mailbox copy vector
 *	    mailbox name
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy_messages (MAILSTREAM *stream, char *mailbox)
  {
    char outmfile[MAILFILESIZE],outfolder[FOLDERSIZE];
    int inbox;
    long i;

    DBGPRINT(21,"vms_mail_copy_messages: entered, stream='%s', destination='%s'", stream->mailbox, mailbox, 0, 0);

    if (__parse_mailbox_name(mailbox, outmfile, outfolder, &inbox) == NIL)
      {
        mm_log("invalid destination folder", ERROR);
        return NIL;
      }

    for (i=1; i <= stream->nmsgs; i++)
      if (mail_elt(stream,i)->sequence)
        if (vms_mail_message_copy(stream,i,outmfile,outfolder) == NIL)
          {
            mm_log("message copy failed", ERROR);
            return NIL;
          }

    DBGPRINT(21,"vms_mail_copy_messages: exit",0,0,0,0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	If this is the INBOX folder and the auto-file inbox flag is set, then
	move all seen messages from the INBOX to the specified mailbox.

   Input:
	stream		The mail stream to be checked

   Output:
	NONE

   Return Value:
	NONE

   Comments:
	This is normally called during a close operation and will normally have
	been preceded by an expunge operation, so some messages flagged as SEEN
	will already have been deleted!

	The LOCAL->inbox flag is used to check if this an inbox

	The logical name IMAPD_AUTOFILEINBOX defines which mailbox the messages
	are to be moved into. If it is not defined, then INBOX messages are not
	moved (leave undefined for standard PINE/IMAPD compatibility, define to
	"MAIL" for VMS MAIL compatibility).

    Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __file_inbox_messages(MAILSTREAM *stream)
  {
    char *nf, mailfile[MAILFILESIZE], folder[FOLDERSIZE];
    int inbox,i,f;
    long filed=0;

    if (!stream) return;
    if (!LOCAL)  return;

    DBGPRINT(21,"__file_inbox_messages: entered, mailbox='%s'", stream->mailbox, 0, 0, 0);


				/* Is this the INBOX ? */
    if (!LOCAL->inbox)
      {
        DBGPRINT(21,"__file_inbox_messages: exit, folder not INBOX", 0, 0, 0, 0);
        return;
      }

				/* Is the autofile option set up? */
    if ((nf = getenv("IMAPD_AUTOFILEINBOX")) == NULL)
      {
        DBGPRINT(21,"__file_inbox_messages: exit, auto-file logical not defined", 0, 0, 0, 0);
        return;
      }

				/* Validate mailbox specification */
    if ( __parse_mailbox_name(nf, mailfile, folder, &inbox) == NIL)
      {
        DBGPRINT(21,"__file_inbox_messages: exit, invalid destination mailbox",0,0,0,0);
        return;
      }


				/* Delete messages SEEN but not DELETED */
    DBGPRINT(25,"__file_inbox_messages: filing seen messages to '%s'", nf, 0, 0, 0);
    for(i = 1; i <= stream->nmsgs; i++)
     {
       f = LOCAL->msgs[i-1].flags;
       if( (f & fSEEN)  &&  !(f & fDELETED) )
        {
					/* Copy the message */
          DBGPRINT(25,"__file_inbox_messages: filing message '%d'", i,0,0,0);
          if (vms_mail_message_copy(stream, i, mailfile, folder) == NIL)
            {
              mm_log("Auto-filing of INBOX messages not completed",WARN);
              return;
            }
					/* Flag it as deleted */
          __internal_setflag(stream, i, mail_elt(stream,i), fDELETED);
          filed++;
        }
     }

					/* Remove messages flagged for delete */
    if (filed)
      { char temp[256];
        sprintf(temp,"auto-filed %d messages to %s", filed, nf);
        mm_log(temp, NIL);
        vms_mail_expunge (stream);
      }    
    DBGPRINT(21,"__file_inbox_messages: exit",0,0,0,0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Open a specific mail file and return the context value

   Input:
	MailFile		The name of the mailfile to open

   Output:
	NONE

   Return Value:
	The callable mail context; or zero if failed

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static int __create_mail_context(char *MailFile)
  {
    int status, context=0;

    item MailFileList[] = {	/* Set the name of the file */
	{ strlen(MailFile), MAIL$_MAILFILE_NAME, MailFile, NULL},
        { 0,                MAIL$_NOSIGNAL,      NULL,     NULL},
	{ 0, 0, NULL, NULL}};

    DBGPRINT(21,"__create_mail_context: Entered.. File='%s'", MailFile,0,0,0);


					/* Set up mailfile context */
    status = mail$mailfile_begin(&context, NullItemList, NullItemList);
    if (!chkstat("mail$mailfile_begin", 0, status)) return 0;
    status = mail$mailfile_open(&context, MailFileList, NullItemList);
    if (!chkstat("mail$mailfile_open", 0, status))
      {
        status = mail$mailfile_end(&context, NullItemList, NullItemList);
        chkstat("mail$mailfile_end", 0, status);
        return 0;
      }


    DBGPRINT(21,"__create_mail_context: exiting, mail ctx=0x%08x", context, 0, 0, 0);
    return context;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Initialize the message context for an opened mail file

   Input:
	mailcontext		Pointer to the mail context for the open file

   Output:
	NONE

   Return Value:
	The message context; or zero if failure

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static int __create_msg_context(int *mailcontext)
  {
    int status, context=0;

    item MessageContext[] = {
	{sizeof(*mailcontext), MAIL$_MESSAGE_FILE_CTX, mailcontext, NULL},
	{0,                    0,                      NULL,        NULL}};

    DBGPRINT(21,"__create_msg_context: entered, mailcontext = '0x%08x'", *mailcontext, 0, 0, 0);
    status = mail$message_begin(&context, MessageContext, NullItemList);
    if (!chkstat("mail$message_begin", 0, status)) return 0;
    DBGPRINT(21,"__create_msg_context: exit, messagecontext='0x%08x'",context,0,0,0);
    return context;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Close the mailfile by ending the specified message and mail contexts

   Inputs:
	mailcontext		Pointer to current mail file context
	messagecontext		Pointer to current message context

   Outputs:
	mailcontext		As above; set to zero
	messagecontext		As above; set to zero

   Return Value:
	NONE

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static void __vms_mail_close(int *mailcontext, int *messagecontext)
  {
    int status;

    item CloseItem[] = {	/* Purge and reclaim during close */
	{ 0, MAIL$_MAILFILE_FULL_CLOSE, NULL, NULL},
	{ 0, 0, NULL, NULL } };

    DBGPRINT(21,"__vms_mail_close: Entered, mail context='0x%08x', messagecontext=0x%08x'",*mailcontext,*messagecontext,0,0);
    if (*mailcontext != 0)
      {
				/* Terminate message context (if any) */
        if (*messagecontext != 0)
          {
            DBGPRINT(25,"__vms_mail_close: Terminating message context (0x%08x)",*messagecontext,0,0,0);
    	    status = mail$message_end(messagecontext, NullItemList, NullItemList);
            chkstat("mail$message_end", 0, status);
          }

				/* Terminate the mail context */
        DBGPRINT(25,"__vms_mail_close: Terminating mail context (0x%08x)",*mailcontext,0,0,0);
        status = mail$mailfile_close(mailcontext, CloseItem, NullItemList);
        chkstat("mail$mailfile_close", 0, status);
        status = mail$mailfile_end(mailcontext, CloseItem, NullItemList);
        chkstat("mail$mailfile_end", 0, status);
      }
    DBGPRINT(21,"__vms_mail_close: exiting, mail ctx=0x%08x, msg ctx=0x%08x", *mailcontext, *messagecontext, 0, 0);
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Test whether a particular mailbox exists by opening the mailfile and
	scanning for the folder.

   Inputs:
	mailbox			The mailbox to check
	mailcontext		Pointer to existing mail context, or NULL
	close			Leave open, or close when done

   Output:
	mailfile		Pointer to parsed mailfile buffer; or NULL
	folder			Pointer to parsed folder buffer; or NULL
	inbox			Pointer to 'is the INBOX' flag
		
   Return Value:
	T			mailbox exists
	NIL			mailbox does not exist

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __folder_exists(char *mailbox, int *mailcontext, int close, char *mailfile, char *folder, int *inbox)
  {
    char tmpmailfile[MAILFILESIZE], tmpfolder[FOLDERSIZE];
    struct stat sbuf;
    struct FolderPattern fp = {0, tmpfolder, NULL};

    int dummycontext = 0;		/* For closing */
    int context      = 0;		/* internal context if none supplied */
    int *ctx         = &context;	/* Context pointer */

    DBGPRINT(21,"__folder_exists: entered, mailbox='%s'",mailbox,0,0,0);

					/* get mailfile and foldername */
    if (!__parse_mailbox_name(mailbox, tmpmailfile, tmpfolder, inbox))
      {
        DBGPRINT(21,"__folder_exists: exit, cannot parse",0,0,0,0);
        return NIL;
      }


    if (stat(tmpmailfile, &sbuf) == -1)
      {
        DBGPRINT(21,"__folder_exists: exit, cannot stat mailfile",0,0,0,0);
        return NIL;
      }

					/* Copy names, if wanted */
    if (mailfile)
      strcpy(mailfile,tmpmailfile);
    if (folder)
      strcpy(folder,tmpfolder);

					/* Use external context if given */
    if (mailcontext)
      ctx = mailcontext;

					/* If not open, open it */
    if (!*ctx)
      {
        DBGPRINT(25,"__folder_exists: opening '%s'", tmpmailfile,0,0,0);
        if ((*ctx = __create_mail_context(tmpmailfile)) == 0)
          {
             DBGPRINT(21,"__folder_exists: exit, open failed",0,0,0,0);
             return NIL;
          }
      }

					/* Scan the 'new' list for a match */
    DBGPRINT(25,"__folder_exists: scanning 'new' list for '%s / %s'",tmpmailfile,tmpfolder,0,0);
    if (__search_new_mailboxes(tmpmailfile,tmpfolder))
       fp.count = 1;			/* Pretend it's been found ! */
    
					/* Scan the folder list for a match */
    DBGPRINT(25,"__folder_exists: scanning '%s' for folder '%s'",tmpmailfile,tmpfolder,0,0);
    __scan_folders(ctx, __find_folder, &fp);


					/* Close if internal or close requested */
    if (context || close)
      {
        DBGPRINT(25,"__folder_exists: closing mail file",0,0,0,0);
        __vms_mail_close(ctx, &dummycontext);
      }

					/* Return T if found, else NIL */
    DBGPRINT(21,"__folder_exists: exit, count ='%d'",fp.count,0,0,0);
    return fp.count ? T : NIL;
  }

/* VMS/MAIL manipulate driver parameters
 * Accepts: function code
 *	    function-dependent value
 * Returns: function-dependent return value
 */

void *vms_mail_parameters (long function, void *value)
{
  return NULL;
}



/* VMS/MAIL mail validate mailbox
 * Accepts: mailbox name
 * Returns: our driver if name is valid, NIL otherwise
 */

DRIVER *vms_mail_valid (char *name)
{
  return vms_mail_isvalid(name) ? &vmsmaildriver : NIL;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Validate the mailbox. Check that the syntax is right for this driver
	and that the amilbox exists.

   Input:
	name		Pointer to mailbox name

   Output:
	NONE

   Return Value:
	T if mailbox exists
	NIL if it doesn't or the format is invalid

  Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
long vms_mail_isvalid (char *name)
{
  static char *invalchrs = "{}()[]#^&'";
  char *i;
  int inbox;

  DBGPRINT(21,"vms_mail_isvalid: entered, name='%s'", name,0,0,0);

						/* Invalid chars ?? */
  for (i=invalchrs; *i; i++)
    if (strchr(name,*i))
      {
        DBGPRINT(21,"vms_mail_isvalid: mailbox name has invalid char '%c'",*i,0,0,0);
        return NIL;
      }

						/* Successful parse ?? */
  if (!__folder_exists(name, NULL, 1, NULL, NULL, &inbox))
    {
     DBGPRINT(21,"vms_mail_isvalid: could not find mailbox '%s'",name,0,0,0);
     return NIL;
    }

  DBGPRINT(21,"vms_mail_isvalid: exiting, mailbox name OK",0,0,0,0);
  return T;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Display all subscribed mailboxes, from the subscription database, which
	match the supplied pattern.

   Inputs:
	stream		The mail stream (not used)
	pat		The pattern to match

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	results are reported back using the mm_mailbox() callback.

	The list of subscribed mail folders is maintained in a database file
	in the user's home directory (see define of MBXDATABASE)

	If the INBOX exists, it is always returned

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
void vms_mail_find (MAILSTREAM *stream, char *pat)
{
  char *home, subscriptionfile[MAILFILESIZE], record[MAXRECORDSIZE];
  FILE *fp;

					/* Always have an INBOX if it exists */
  if (matchwild(pat,"INBOX") && vms_mail_isvalid("INBOX"))
    mm_mailbox("INBOX");

					/* Find location of subscription database */
  if (__get_home_directory(&home) == NULL)
    {
      mm_log("cannot locate subscription database",ERROR);
      return;
    }

					/* Scan subscription database */
  strcpy(subscriptionfile,home);
  strcat(subscriptionfile,MBXDATABASE);
  if ((fp=fopen(subscriptionfile,"r")) != NULL)
    {
      while ( fgets(record, MAXRECORDSIZE, fp)  )
       {
        record[strlen(record)-1] = '\0';/* strip '\n' */
        if ( matchwild(pat, record) )
           mm_mailbox(record);
       }
      fclose(fp);
    }

  return;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Find all the mailboxes that match the pattern. This actually returns
	all folders in the mail file(s) that match the pattern.

   Inputs:
	stream		The mail stream (not used)
	pat		The pattern to match

   Outputs:
	NONE

   Return Value:
	NONE

   Comments:
	Results are reported back via the mm_mailbox() callback.

	The nominated mail file is scanned and all matching folders are
	reported.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
void vms_mail_find_all (MAILSTREAM *stream, char *pat)
{
  char *p, prefix[MAILFILESIZE], mailfile[MAILFILESIZE];
  int context=0, dummymessagecontext=0, inbox;
  struct FolderPattern fp = {0, pat, ""};

  DBGPRINT(21,"vms_mail_find_all: entered, pattern = '%s'", pat,0,0,0);


					/* Find possible mailfile name prefix */
   strcpy(prefix, pat);
   if ((p = strrchr(prefix,'/')) != NULL)
     {
       *(p+1) = '\0';
       DBGPRINT(25,"vms_mail_find_all: found mailbox prefix '%s'", prefix, 0, 0, 0);
       fp.data = prefix;
     }

					/* parse mailbox pattern */
  if (__parse_mailbox_name(pat, mailfile, NULL, &inbox) == NIL)
    {
       mm_log("Invalid folder pattern", ERROR);
       DBGPRINT(21,"vms_mail_find_all: exit : invalid pattern",0,0,0,0);
       return;
    }

					/* open the mail file */
  if ( (context=__create_mail_context(mailfile)) == 0 )
    {
      DBGPRINT(21,"vms_mail_find_all: exit : cannot open mail file '%s'", mailfile,0,0,0);
      return;
    }

					/* Find matching folders */
  __scan_folders(&context, __match_folders, &fp);
  __vms_mail_close(&context,&dummymessagecontext);	/* Close it */

  DBGPRINT(21,"vms_mail_find_all: exit",0,0,0,0);
}

/* VMS/MAIL mail find list of bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_bboards (MAILSTREAM *stream, char *pat)
{
  return;		/* no bboards supported on VMS */
}



/* VMS/MAIL mail find list of all bboards
 * Accepts: mail stream
 *	    pattern to search
 */

void vms_mail_find_all_bboards (MAILSTREAM *stream,char *pat)
{
   return;		/* Again, not supported at this stage */
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Add a new mailbox to the list of subscribed folders in the subscription
	database.

   Inputs:
	stream		The mail stream (not used)
	mailbox		Name of mailbox to add

   Outputs:
	NONE

   Return Value:
	T		If subscription accepted
	NIL		If subscription rejected

   Comments:
	Since INBOX is automatically included on the search of the subscription
	database, it may not be added to the file.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_subscribe (MAILSTREAM *stream, char *mailbox)
{
  char *home,subscriptionlist[MAILFILESIZE], record[MAXRECORDSIZE];
  FILE *fp;

					/* INBOX has special treatment */
  if (strcasecmp(mailbox, "INBOX"))
    {
      mm_log("INBOX already implictly subscribed",ERROR);
      return NIL;
    }

					/* Get home directory */
  if (__get_home_directory(&home) == NULL)
    {
       mm_log("can't locate subscription database",ERROR);
       return NIL;
    }

					/* Construct subscription database name */
  strcpy(subscriptionlist, home);
  strcat(subscriptionlist,MBXDATABASE);


					/* Do we have a subscription database ?? */
  if ((fp=fopen(subscriptionlist,"r")) != NULL)
    {
					/* Is mailbox already subscribed? */
      while ( fgets(record, MAXRECORDSIZE, fp)  )
        {
          record[strlen(record)-1] = '\0';	/* remove '\n' */
          if ( strcasecmp(mailbox,record) )
            {
              fclose(fp);
              sprintf(record,"already subscribed to %s",mailbox);
              mm_log(record,ERROR);
              return NIL;
            }
        };
      fclose(fp);
    }


						/* Append new subscription */
   if ((fp=fopen(subscriptionlist,"a","rfm=var","rat=cr")) == NULL)
     {
       mm_log("unable to append to subscription database",ERROR);
       return NIL;
     }
   fputs(lcase(mailbox),fp); fputc('\n',fp);
   fclose(fp);
   return T;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Delete a mailbox from the mailbox subscription database

   Input:
	stream		Mail stream (not used)
	mailbox		Mailbox to unsubscribe

   Output:
	NONE

   Return Value:
	T		Subscription successfully removed
	NIL		Subscription not removed

   Comments:
	The name is removed from the subscription file in the user's home
        directory.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_unsubscribe (MAILSTREAM *stream, char *mailbox)
{
  char subscriptionfile[MAILFILESIZE], newsubscriptionfile[MAILFILESIZE], oldsubscriptionfile[MAILFILESIZE];
  char *home, record[MAXRECORDSIZE];
  FILE *fp,*newfp;
  int subscribed;

					/* INBOX not valid */
  if ( strcasecmp("INBOX", mailbox) )
    {
      mm_log("cannot unsubscribe INBOX",ERROR);
      return NIL;
    }

					/* Get location of subscription database */
  if (__get_home_directory(&home) == NULL)
    {
      mm_log("cannot locate subscription database",ERROR);
      return NIL;
    }

					/* Construct subscription database names */
  strcpy(subscriptionfile, home);
  strcat(subscriptionfile, MBXDATABASE);
  
  strcpy(oldsubscriptionfile, subscriptionfile);
  strcat(oldsubscriptionfile, "-OLD");
  strcpy(newsubscriptionfile, subscriptionfile);
  strcat(newsubscriptionfile, "-NEW");


					/* Open existing subscriptions */
  if ((fp=fopen(subscriptionfile,"r")) == NULL)
    {
      mm_log("no subscription file", ERROR);
      return NIL;
    }


					/* Create new subscription database */
  if ((newfp=fopen(newsubscriptionfile,"w","rfm=var","rat=cr")) == NULL)
    {
      fclose(fp);
      mm_log("cannot create new subscription database",ERROR);
      return NIL;
    }
      
					/* Copy, excluding one to delete */
  subscribed = 0;
  while ( fgets(record, MAXRECORDSIZE, fp) )
    {
       record[strlen(record)-1] = '\0';

       if (strcasecmp(mailbox,record))
	  { subscribed++; }
       else
          { fputs(record,newfp); fputc('\n',newfp);}
    }
  fclose(fp);
  fclose(newfp);


					/* Did we find the mailbox ?? */
  if (!subscribed)
    {
      remove(newsubscriptionfile);
      mm_log("mailbox not subscribed",ERROR);
      return NIL;
    }


					/* Preserve old subscriptions */
  if (rename(subscriptionfile, oldsubscriptionfile) != 0)
    {
      remove(newsubscriptionfile);
      mm_log("cannot save old subscription database",ERROR);
      return NIL;
    }


					/* Rename new copy to master */
  if (rename(newsubscriptionfile, subscriptionfile) != 0)
    {
       remove(newsubscriptionfile);
       rename(oldsubscriptionfile,subscriptionfile);
       mm_log("cannot rename new subscription database",ERROR);
       return NIL;
    }

					/* Get rid of old copy */
  if (remove(oldsubscriptionfile) !=0 )
     mm_log("could not remove old subscription file", WARN);

  return T;
}

/* VMS/MAIL mail subscribe to bboard
 * Accepts: mail stream
 *	    bboard to add to subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_subscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  mm_log("not implemented on VMS",ERROR);
  return NIL;			/* never valid for VMS/MAIL */
}


/* VMS/MAIL mail unsubscribe to bboard
 * Accepts: mail stream
 *	    bboard to delete from subscription list
 * Returns: T on success, NIL on failure
 */

long vms_mail_unsubscribe_bboard (MAILSTREAM *stream, char *mailbox)
{
  mm_log("not implemented on VMS",ERROR);
  return NIL;			/* never valid for VMS/MAIL */
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Create a new mailbox

   Inputs:
	stream		Name of mail stream (not used)
	mailbox		Name of mailbox to create

   Output:
	NONE

   Return Value:
	T		mailbox successfully created
	NIL		mailbox not created

   Comments:
	This routine creates an empty file based on the FDL specification of
	another file:
		Either: the user's default MAIL file
		Or:	The prototype specified by IMAPD_MAILPROTYPE

	We cannot create a 'folder' in a VMS MAIL file so we only create the
	mail file itself. VMS MAIL does not allow empty folders to exist as
	they are created automatically when a message is copied into them.

	We need to handle the checking carefully. We must check that the named
	folder exists within the mail file.

   Author:
	Andy Harper, Kings College London, Jan 1997

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */

long vms_mail_create (MAILSTREAM *stream, char *mailbox)
{
   char mailfile[MAILFILESIZE], folder[FOLDERSIZE], fdlbuffer[4096];
   char tmp[256], maildefault[MAILFILESIZE];
   char *p;
   int fdlsize,i,inbox;
   struct stat sbuf;

   char *prototypes[] = {maildefault,"IMAPD_MAILPROTOTYPE"};
   int maxproto       = sizeof(prototypes)/sizeof(prototypes[0]);
   int first          = 0;
   int context        = 0;
   int dummycontext   = 0;


   DBGPRINT(21,"vms_mail_create: entered, mailbox='%s'", mailbox,0,0,0);

   if (!__parse_mailbox_name("INBOX", maildefault, NULL, &inbox))
     {
       DBGPRINT(25,"vms_mail_create: can't parse INBOX",0,0,0,0);
       first++;	/* skip attempt to get prototype from INBOX */
     }
      

   DBGPRINT(25,"vms_mail_create: checking existence of folder",0,0,0,0);
   if (__folder_exists(mailbox, &context, 1, NULL, NULL, &inbox))
     {
       mm_log("folder already exists",ERROR);
       return NIL;
     }

   DBGPRINT(25,"vms_mail_create: parsing mailbox",0,0,0,0)
   if (!__parse_mailbox_name(mailbox,mailfile,folder,&inbox))
     {
       mm_log("unparseable mailbox name",ERROR);
       return NIL;
     }

   DBGPRINT(25,"vms_mail_create: checking existence of mail file",0,0,0,0);
   if (stat(mailfile, &sbuf) == 0)
     {
       DBGPRINT(25,"vms_mail_create: mailfile exists, but no folder",0,0,0,0);
       __Add_New_Mailbox(mailfile,folder);
       return T;
     }
  
   
   DBGPRINT(25,"vms_mail_create: creating directory", 0, 0, 0, 0);
   if ((p=strchr(mailfile,']')) != NULL)
     {
       strncpy(tmp, mailfile, (p-mailfile)+1);
       DBGPRINT(25,"vms_mail_create: directory is '%s'", tmp,0,0,0);
       if (mkdir(tmp,0) != 0)
         {
           DBGPRINT(25,"vms_mail_create: directory create failed",0,0,0,0);
         }
     }

   DBGPRINT(25,"vms_mail_create: attempting create from prototype",0,0,0,0);
   for (i=first; i<maxproto; i++)
     {
       if (!__fdl_generate(prototypes[i], fdlbuffer, 
sizeof(fdlbuffer)-1, &fdlsize)) continue;
       if (!__fdl_create_file(mailfile, fdlbuffer, fdlsize)) continue;
       __Add_New_Mailbox(mailfile,folder);
       DBGPRINT(21,"vms_mail_create: exit, created mailfile",0,0,0,0);
       return T;
     }

   sprintf(tmp,"could not create '%s'",mailbox);
   mm_log(tmp, ERROR);
   DBGPRINT(21,"vms_mail_create: exit, create failed",0,0,0,0);
   return NIL;
}

/* VMS/MAIL mail delete mailbox
 * Accepts: MAIL stream
 *	    mailbox name to delete
 * Returns: T on success, NIL on failure
 */

long vms_mail_delete (MAILSTREAM *stream, char *mailbox)
{
   DBGPRINT(21,"vms_mail_delete: entered, mailbox='%s'", mailbox,0,0,0)
   mm_log("not supported on VMS", ERROR);
   DBGPRINT(21,"vms_mail_delete: exit",0,0,0,0);
  return NIL;
}


/* VMS/MAIL mail rename mailbox
 * Accepts: MAIL stream
 *	    old mailbox name
 *	    new mailbox name (or NIL for delete)
 * Returns: T on success, NIL on failure
 */

long vms_mail_rename (MAILSTREAM *stream, char *old, char *new)
{
   DBGPRINT(21,"vms_mail_rename: entered, old='%s', new='%s'", old,new,0,0)
   mm_log("not supported on VMS", ERROR);
   DBGPRINT(21,"vms_mail_rename: exit",0,0,0,0);
   return NIL;
}

/* VMS/MAIL mail open - open a "mailbox" which is a folder in VMS/MAIL.
 * Accepts: stream to open
 * Returns: stream on success, NIL on failure
 */

MAILSTREAM *vms_mail_open (MAILSTREAM *stream)
{
  int MailContext=0, MessageContext=0;
  char mailfile[MAILFILESIZE], folder[FOLDERSIZE];
  int inbox, status;
  long i;

					/* return prototype for OP_PROTOTYPE call */
  if (!stream) return &vmsmailproto;


  DBGPRINT(21,"vms_mail_open: Entered, mailbox='%s'",stream->mailbox,0,0,0);


  if (!__folder_exists(stream->mailbox, &MailContext, 0, mailfile, folder, &inbox))
    {
      DBGPRINT(25,"vms_mail_open: folder not found",0,0,0,0);
      mm_log("no such mailbox",ERROR);
      return NIL;
    }
					/* close old file if stream being recycled */
  if (LOCAL) {
    vms_mail_close (stream);		/* dump and save the changes */
    stream->dtb = &vmsmaildriver;	/* reattach this driver */
    mail_free_cache (stream);		/* clean up cache */
  }


					/* Create folder and message context */
  if ((MessageContext = __create_msg_context(&MailContext)) == 0)
   {
      __vms_mail_close(&MailContext, &MessageContext);
      DBGPRINT(21,"vms_mail_open: exit, status = 0 : can't create msg context",0,0,0,0);
      mm_log("cannot open folder", ERROR);
      return NIL;
   }


  stream->dtb = &vmsmaildriver;		/* reattach this driver */
  stream->local = fs_get (sizeof (VMSMAILLOCAL));


					/* Save VMS format of the mailbox */
  strcpy(LOCAL->mailfile,mailfile);
  strcpy(LOCAL->folder,folder);

					/* Note if it is an INBOX */
  LOCAL->inbox=inbox;

					/* Save the mail and message contexts */
  LOCAL->mail_context = MailContext;
  LOCAL->msg_context  = MessageContext;


					/* Provide buffer space */
  LOCAL->buflen = INITTEXTSIZE;
  LOCAL->buf = (char *) fs_get (INITTEXTSIZE + 1);


					/* force readonly if bboard */
  if (*stream->mailbox == '*') stream->rdonly = T;
  stream->sequence++;			/* bump sequence number */


					/* parse mailbox */

  LOCAL->pinged = 0;			/* not 'ping'ed yet! */
  stream->nmsgs = stream->recent = 0;
  if (vms_mail_ping (stream) && !stream->nmsgs)
	mm_log ("Mailbox is empty",(long) NIL);

					/* Initialize the local message cache */
  __init_local_cache(stream);

					/* All elts have valid flags ??? */
  for (i = 1; i < stream->nmsgs; i++)
    mail_elt (stream,i)->valid = T;

					/* Notify application of msg count */
  mail_exists(stream,stream->nmsgs);

  DBGPRINT(21,"vms_mail_open: exiting",0,0,0,0);
  return stream;			/* return stream to caller */

}

/* VMS/MAIL mail close
 * Accepts: MAIL stream
 */

void vms_mail_close (MAILSTREAM *stream)
{
  DBGPRINT(21,"vms_mail_close: entered, mailbox='%s'", stream->mailbox,0,0,0);

					/* if the stream is open, close it */
  if (stream)
    {
      DBGPRINT(21,"vms_mail_close: closing stream",0,0,0,0);
      if (LOCAL)
        {
          __file_inbox_messages(stream);		/* auto-file read msg*/
          __clear_local_cache(stream);			/* clear cache */
          __vms_mail_close(&LOCAL->mail_context,&LOCAL->msg_context);	/* close file */

							/* return local data*/
          if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
          fs_give ((void **) &stream->local);
        }
      stream->dtb = NIL;				/* log out the DTB */
    }

  DBGPRINT(21,"vms_mail_close: exiting",0,0,0,0);
}

/* VMS/MAIL mail fetch fast information
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchfast (MAILSTREAM *stream, char *sequence)
{
  long i;

  DBGPRINT(21,"vms_mail_fetchfast: mailbox='%s', sequence='%s'", stream->mailbox, sequence, 0, 0);
				/* scan the sequence */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence)
         vms_mail_fetchstructure(stream,i,NIL);

  DBGPRINT(21,"vms_mail_fetchfast: exit",0,0,0,0);
  return;
}


/* VMS/MAIL mail fetch flags
 * Accepts: MAIL stream
 *	    sequence
 */

void vms_mail_fetchflags (MAILSTREAM *stream, char *sequence)
{
  long i;

  DBGPRINT(21,"vms_mail_fetchflags: mailbox='%s', sequence='%s'", stream->mailbox, sequence, 0, 0);
				/* ping mailbox, get new status for messages */
  if (mail_sequence (stream,sequence))
    for (i = 1; i <= stream->nmsgs; i++)
      if (mail_elt (stream,i)->sequence) {
	MESSAGECACHE *elt = mail_elt (stream,i);
        FILECACHE    *fc  = &LOCAL->msgs[i-1];
	elt->deleted  = ((fc->flags & fDELETED)  ? T : NIL);
	elt->seen     = ((fc->flags & fSEEN)     ? T : NIL);
	elt->flagged  = ((fc->flags & fFLAGGED)  ? T : NIL);
	elt->answered = ((fc->flags & fANSWERED) ? T : NIL);
     }
  DBGPRINT(21,"vms_mail_fetchflags: exit",0,0,0,0);
}

/* VMS/MAIL mail fetch structure
 * Accepts: MAIL stream
 *	    message # to fetch
 *	    pointer to return body
 * Returns: envelope of this message, body returned in body value
 *
 * Fetches the "fast" information as well
 */
ENVELOPE *vms_mail_fetchstructure (MAILSTREAM *stream, long msgno, BODY **body)
{
  FILECACHE *fc = &LOCAL->msgs[msgno-1];
  LONGCACHE *lelt;
  ENVELOPE **env;
  BODY **b;
  STRING bs;

  DBGPRINT(21,"vms_mail_fetchstructure: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);
  if (stream->scache) {		/* short cache */
    if (msgno != stream->msgno){/* flush old poop if a different message */
      mail_free_envelope (&stream->env);
      mail_free_body (&stream->body);
    }
    stream->msgno = msgno;
    env = &stream->env;		/* get pointers to envelope and body */
    b = &stream->body;
  }
  else {			/* long cache */
    lelt = mail_lelt (stream,msgno);
    env = &lelt->env;		/* get pointers to envelope and body */
    b = &lelt->body;
  }
  if ((body && !*b) || !*env) {	/* have the poop we need? */
    mail_free_envelope (env);	/* flush old envelope and body */
    mail_free_body (b);

				/* Fetch message headers and body */
    __read_message(stream,msgno);

    INIT (&bs,mail_string,(void *) fc->text,fc->size);
				/* parse envelope and body */
    rfc822_parse_msg (env,body ? b : NIL,fc->header,fc->headersize,&bs,mylocalhost(),LOCAL->buf);
    (lelt->elt).rfc822_size = fc->headersize + fc->size;       
  }

  DBGPRINT(21,"vms_mail_fetchstructure: exit",0,0,0,0);
  if (body) *body = *b;		/* return the body */
  return *env;			/* return the envelope */
}

/* VMS/MAIL mail fetch message header
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message header in RFC822 format
 */

char *vms_mail_fetchheader (MAILSTREAM *stream,long msgno)
{
  FILECACHE *fc = &LOCAL->msgs[msgno-1];

  DBGPRINT(21,"vms_mail_fetchheader: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);

  __read_message(stream,msgno);
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,fc->header,fc->headersize);

  DBGPRINT(21,"vms_mail_fetchheader: exit",0,0,0,0);
  return LOCAL->buf;
}




/* VMS/MAIL mail fetch message text (only)
	body only;
 * Accepts: MAIL stream
 *	    message # to fetch
 * Returns: message text in RFC822 format
 */

char *vms_mail_fetchtext (MAILSTREAM *stream, long msgno)
{
  FILECACHE    *fc  = &LOCAL->msgs[msgno-1];
  MESSAGECACHE *elt = mail_elt(stream,msgno);

  DBGPRINT(21,"vms_mail_fetchtext: mailbox='%s', msgno='%d'", stream->mailbox, msgno, 0, 0);
				/* get to text position */
  __read_message(stream,msgno);
				/* copy the string */
  strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,fc->text,fc->size);

				/* recalculate status */
  if (!elt->seen)
    { elt->seen = T; __update_message_status (stream,msgno,T); }

  DBGPRINT(21,"vms_mail_fetchtext: exit",0,0,0,0);
  return LOCAL->buf;
}

/* VMS/MAIL fetch message body as a structure
 * Accepts: Mail stream
 *	    message # to fetch
 *	    section specifier
 *	    pointer to length
 * Returns: pointer to section of message body
 */
char *vms_mail_fetchbody (MAILSTREAM *stream, long m, char *s, unsigned long *len)
{
  BODY *b;
  PART *pt;
  char *t;
  ENVELOPE *env;
  unsigned long i;
  unsigned long offset = 0;
  MESSAGECACHE *elt = mail_elt (stream,m);
  FILECACHE    *fc  = &LOCAL->msgs[m-1];

  DBGPRINT(21,"vms_mail_fetchbody: mailbox='%s', msgno='%d', section='%s'", stream->mailbox, m, s, 0);
				/* make sure have a body */
  env = vms_mail_fetchstructure (stream,m,&b);
  if (!(env && b && s && *s &&
	((i = strtol (s,&s,10)) > 0))) return NIL;
  do {				/* until find desired body part */
				/* multipart content? */
    if (b->type == TYPEMULTIPART) {
      pt = b->contents.part;	/* yes, find desired part */
      while (--i && (pt = pt->next));
      if (!pt) return NIL;	/* bad specifier */
				/* note new body, check valid nesting */
      if (((b = &pt->body)->type == TYPEMULTIPART) && !*s) return NIL;
      offset = pt->offset;	/* get new offset */
    }
    else if (i != 1) return NIL;/* otherwise must be section 1 */
				/* need to go down further? */
    if (i = *s) switch (b->type) {
    case TYPEMESSAGE:		/* embedded message, calculate new base */
      offset = b->contents.msg.offset;
      b = b->contents.msg.body;	/* get its body, drop into multipart case */
    case TYPEMULTIPART:		/* multipart, get next section */
      if ((*s++ == '.') && (i = strtol (s,&s,10)) > 0) break;
    default:			/* bogus subpart specification */
      return NIL;
    }
  } while (i);
				/* lose if body bogus */
  if ((!b) || b->type == TYPEMULTIPART) return NIL;

  if (!elt->seen)
    { elt->seen=T;  __update_message_status(stream,m,T); }

				/* move to that place in the data */
  t = fc->text + offset;

  rfc822_contents(&LOCAL->buf,&LOCAL->buflen,len,t,b->size.ibytes,b->encoding);
  DBGPRINT(21,"vms_mail_fetchbody: exit",0,0,0,0);
  return LOCAL->buf;
}

/* Parse flag list
 * Accepts: MAIL stream
 *	    flag list as a character string
 * Returns: flag command list
 */


short vms_mail_getflags (MAILSTREAM *stream, char *flag)
{
  char tmpbuf[256];
  char *t;
  short f = 0;
  short i,j;
  if (flag && *flag) {		/* no-op if no flag string */
				/* check if a list and make sure valid */
    if ((i = (*flag == '(')) ^ (flag[strlen (flag)-1] == ')')) {
      mm_log ("Bad flag list",ERROR);
      return NIL;
    }
				/* copy the flag string w/o list construct */
    strncpy (tmpbuf,flag+i,(j = strlen (flag) - (2*i)));
    tmpbuf[j] = '\0';
    t = ucase (tmpbuf);		/* uppercase only from now on */

    while (*t) {		/* parse the flags */
      if (*t == '\\') {		/* system flag? */
	switch (*++t) {		/* dispatch based on first character */
	case 'S':		/* possible \Seen flag */
          if (strncmp(t+1, "EEN", 3) == 0) i = fSEEN;
	  t += 4;		/* skip past flag name */
	  break;
	case 'D':		/* possible \Deleted flag */
          if (strncmp(t+1, "ELETED", 6) == 0) i = fDELETED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'F':		/* possible \Flagged flag */
          if (strncmp(t+1, "LAGGED", 6) == 0) i = fFLAGGED;
	  t += 7;		/* skip past flag name */
	  break;
	case 'A':		/* possible \Answered flag */
          if (strncmp(t+1,"NSWERED", 7) == 0) i = fANSWERED;
	  t += 8;		/* skip past flag name */
	  break;
	default:		/* unknown */
	  i = 0;
	  break;
	}
				/* add flag to flags list */
	if (i && ((*t == '\0') || (*t++ == ' '))) f |= i;
	else {			/* bitch about bogus flag */
	  mm_log ("Unknown system flag",ERROR);
	  return NIL;
	}
      }
      else {			/* no user flags yet */
	mm_log ("Unknown flag",ERROR);
	return NIL;
      }
    }
  }
  return f;
}

/* VMS/MAIL mail set flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_setflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */

				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence)
      {
        __internal_setflag(stream,i,elt,f);
        __update_message_status(stream,i,T);
      }
}




/* VMS/MAIL mail clear flag
 * Accepts: MAIL stream
 *	    sequence
 *	    flag(s)
 */

void vms_mail_clearflag (MAILSTREAM *stream, char *sequence, char *flag)
{
  int	i;
  MESSAGECACHE *elt;

  short f = vms_mail_getflags (stream,flag);	/* Parse the list */
  if (!f) return;		/* no-op if no flags to modify */
				/* get sequence and loop on it */
  if (mail_sequence (stream,sequence)) for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence)
      {
        __internal_clearflag(stream,i, elt, f);
        __update_message_status(stream,i,T);
      }
}

/* VMS/MAIL mail search for messages
 * Accepts: MAIL stream
 *	    search criteria
 */

void vms_mail_search (MAILSTREAM *stream, char *criteria)
{
  long i,n;
  char *d;
  search_t f;
				/* make sure all messages read in */
  for (i=1; i<=stream->nmsgs; i++)
    vms_mail_fetchstructure(stream, i, NIL);

				/* initially all searched */
  for (i = 1; i <= stream->nmsgs; ++i) mail_elt (stream,i)->searched = T;
				/* get first criterion */
  if (criteria && (criteria = strtok (criteria," "))) {
				/* for each criterion */
    for (; criteria; (criteria = strtok (NIL," "))) {
      f = NIL; d = NIL; n = 0;	/* init then scan the criterion */
      switch (*ucase (criteria)) {
      case 'A':			/* possible ALL, ANSWERED */
	if (!strcmp (criteria+1,"LL")) f = vms_mail_search_all;
	else if (!strcmp (criteria+1,"NSWERED")) f = vms_mail_search_answered;
	break;
      case 'B':			/* possible BCC, BEFORE, BODY */
	if (!strcmp (criteria+1,"CC"))
	  f = vms_mail_search_string (vms_mail_search_bcc,&d,&n);
	else if (!strcmp (criteria+1,"EFORE"))
	  f = vms_mail_search_date (vms_mail_search_before,&n);
	else if (!strcmp (criteria+1,"ODY"))
	  f = vms_mail_search_string (vms_mail_search_body,&d,&n);
	break;
      case 'C':			/* possible CC */
	if (!strcmp (criteria+1,"C"))
	  f = vms_mail_search_string (vms_mail_search_cc,&d,&n);
	break;
      case 'D':			/* possible DELETED */
	if (!strcmp (criteria+1,"ELETED")) f = vms_mail_search_deleted;
	break;
      case 'F':			/* possible FLAGGED, FROM */
	if (!strcmp (criteria+1,"LAGGED")) f = vms_mail_search_flagged;
	else if (!strcmp (criteria+1,"ROM"))
	  f = vms_mail_search_string (vms_mail_search_from,&d,&n);
	break;
      case 'K':			/* possible KEYWORD */
	if (!strcmp (criteria+1,"EYWORD"))
	  f = vms_mail_search_flag (vms_mail_search_keyword,&n,stream);
	break;
      case 'N':			/* possible NEW */
	if (!strcmp (criteria+1,"EW")) f = vms_mail_search_new;
	break;
      case 'O':			/* possible OLD, ON */
	if (!strcmp (criteria+1,"LD")) f = vms_mail_search_old;
	else if (!strcmp (criteria+1,"N"))
	  f = vms_mail_search_date (vms_mail_search_on,&n);
	break;
      case 'R':			/* possible RECENT */
	if (!strcmp (criteria+1,"ECENT")) f = vms_mail_search_recent;
	break;
      case 'S':			/* possible SEEN, SINCE, SUBJECT */
	if (!strcmp (criteria+1,"EEN")) f = vms_mail_search_seen;
	else if (!strcmp (criteria+1,"INCE"))
	  f = vms_mail_search_date (vms_mail_search_since,&n);
	else if (!strcmp (criteria+1,"UBJECT"))
	  f = vms_mail_search_string (vms_mail_search_subject,&d,&n);
	break;
      case 'T':			/* possible TEXT, TO */
	if (!strcmp (criteria+1,"EXT"))
	  f = vms_mail_search_string (vms_mail_search_text,&d,&n);
	else if (!strcmp (criteria+1,"O"))
	  f = vms_mail_search_string (vms_mail_search_to,&d,&n);
	break;
      case 'U':			/* possible UN* */
	if (criteria[1] == 'N') {
	  if (!strcmp (criteria+2,"ANSWERED")) f = vms_mail_search_unanswered;
	  else if (!strcmp (criteria+2,"DELETED")) f = vms_mail_search_undeleted;
	  else if (!strcmp (criteria+2,"FLAGGED")) f = vms_mail_search_unflagged;
	  else if (!strcmp (criteria+2,"KEYWORD"))
	    f = vms_mail_search_flag (vms_mail_search_unkeyword,&n,stream);
	  else if (!strcmp (criteria+2,"SEEN")) f = vms_mail_search_unseen;
	}
	break;
      default:			/* we will barf below */
	break;
      }
      if (!f) {			/* if can't determine any criteria */
	sprintf (LOCAL->buf,"Unknown search criterion: %.80s",criteria);
	mm_log (LOCAL->buf,ERROR);
	return;
      }
				/* run the search criterion */
      for (i = 1; i <= stream->nmsgs; ++i)
	if (mail_elt (stream,i)->searched && !(*f) (stream,i,d,n))
	  mail_elt (stream,i)->searched = NIL;
    }
				/* report search results to main program */
    for (i = 1; i <= stream->nmsgs; ++i)
      if (mail_elt (stream,i)->searched) mail_searched (stream,i);
  }
}

/* VMS/MAIL mail ping mailbox - find how many messages are in this folder.
 * Do it only when opening the folder.
 *
 * Accepts: MAIL stream
 * Returns: T if stream still alive, NIL if not
 */
long vms_mail_ping (MAILSTREAM *stream)
{
  static unsigned int MessagesCount;
  int status;

  item FolderItem[] = {				/* Folder's name */
	{ 0, MAIL$_MESSAGE_FOLDER, NULL, NULL },/* Must be the first item!!! */
	{ 0, MAIL$_NOSIGNAL,       NULL, NULL},
	{ 0, 0, NULL, NULL }};

  item MessagesCountItem[] = {			/* No of messages */
	{ sizeof(MessagesCount), MAIL$_MESSAGE_SELECTED,&MessagesCount, NULL },
	{ 0, 0, NULL, NULL }};

  DBGPRINT(21,"vms_mail_ping: entered, mailbox='%s'",stream->mailbox,0,0,0);

  if (!LOCAL->pinged) 		/* Not checked yet ? */
    {
      FolderItem[0].buffer = LOCAL->folder;
      FolderItem[0].length = strlen(LOCAL->folder);
      status = mail$message_select(&LOCAL->msg_context, FolderItem, MessagesCountItem);
      if (!chkstat("mail$message_select", 0, status) ) return NIL;

      LOCAL->pinged = 1;
      stream->nmsgs = MessagesCount;
    }

  DBGPRINT(21,"vms_mail_ping: exit, messages='%u'", MessagesCount,0,0,0);
  return T;		/* local stream is always alive */
}

/* VMS/MAIL mail check mailbox (too)
	reparses status too;
 * Accepts: MAIL stream
 */

void vms_mail_check (MAILSTREAM *stream)
{
  mm_log ("Check completed",(long) NIL);
}

/* VMS/MAIL mail expunge mailbox - Delete the messages that are flagged as
 * deleted.
 * Accepts: MAIL stream
 * 
 *
 * Comments:
 *      Expunge may be called twice for the same stream. Once just prior
 *      to calling the close operation and again, if an INBOX, as part of the
 *      close operation. The latter is used to delete messages that have been
 *      auto-filed to another folder.
 * 
 *	To avoid problems in deleting messages twice, we use a separate flag
 *	WAS_DELETED to indicate if we deleted the message the first time
 *	around. That way, we know not to delete it again.
 */

void vms_mail_expunge (MAILSTREAM *stream)
{
  static int MessageNum;		/* Message number to delete */
  int status;
  long i, f;
  long MessagesDeleted = 0;		/* How many messages were deleted */

  item DeleteMessageItem[] = {
	{ sizeof(MessageNum), MAIL$_MESSAGE_ID, &MessageNum, NULL },
	{ 0, 0, NULL, NULL }};

  
  DBGPRINT(21,"vms_mail_expunge: entered, mailbox='%s'", stream->mailbox,0,0,0);
  for (i = 1; i <= stream->nmsgs; ++i)
    {
      f = LOCAL->msgs[i-1].flags;
      if( (f & fDELETED)  &&  !(f & fWAS_DELETED) )
	{
	  MessageNum = i;	/* Assign it at the descriptor */
	  status = mail$message_delete(&LOCAL->msg_context, &DeleteMessageItem, NullItemList);
	  if (chkstat("mail$message_delete", 0, status))
            {
              DBGPRINT(25,"vms_mail_expunge: Message '%d' deleted", i, 0, 0, 0);
	      LOCAL->msgs[i - 1].flags |= fWAS_DELETED;
              MessagesDeleted++;	/* Count ALL deleted messages */
            }
	}
    }


  if (MessagesDeleted)			/* If any messages were removed */
    {
     char expmess[256];
     sprintf(expmess,
       (MessagesDeleted==1) ?
          "Expunged %d message" :
          "Expunged %d messages",    MessagesDeleted);
     mm_log(expmess,NIL);     
    }

  DBGPRINT(21,"vms_mail_expunge: exit, messages expunged ='%d'", MessagesDeleted,0,0,0);
}

/* VMS/MAIL mail copy message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_copy (MAILSTREAM *stream, char *sequence, char *mailbox)
{
				/* copy the messages */
  return (mail_sequence (stream,sequence)) ?
    vms_mail_copy_messages (stream,mailbox) : NIL;
}


/* VMS/MAIL mail move message(s)
 * Accepts: MAIL stream
 *	    sequence
 *	    destination mailbox
 * Returns: T if success, NIL if failed
 */

long vms_mail_move (MAILSTREAM *stream, char *sequence, char *mailbox)
{
  long i;
  MESSAGECACHE *elt;
  if (!(mail_sequence (stream,sequence) &&
	vms_mail_copy_messages (stream,mailbox))) return NIL;

				/* delete all requested messages */
  for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence) {
      elt->deleted = T;		/* mark message deleted */
				/* recalculate status */
      __update_message_status (stream,i,NIL);
    }
				/* make sure the update takes */
  return LONGT;
}

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Copy message(s) from the named sequential mail file into another
	mailfile/folder

    Inputs:
	seqfile		Name of a sequential mail file containing source
	msg_count	Number of messages to copy
	mailfile	Name of destination indexed mail file
	folder		Name of destination mail folder

    Outputs:
	NONE

    Return Status:
	T		Copied OK
	NIL		Some problem during copy (noted via mm_log)

    Author:
	Thanks to Steve Pulley of Digital for basic idea and code, November 1996
	Modified by Andy Harper, Kings College London, November 1996.

 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/

static int __copy_sequential_mail_messages(char *seqfile, int msg_count, char *mailfile, char *folder)
  {
    static int id, newfile, newfold;
    int file_ctx=0, msg_context=0;
    int status;

    item copy_in[]   = {
	{sizeof(id),       MAIL$_MESSAGE_ID,               &id,        NULL},
	{strlen(mailfile), MAIL$_MESSAGE_FILENAME,         mailfile,   NULL},
	{strlen(folder),   MAIL$_MESSAGE_FOLDER,           folder,     NULL},
	{0,                MAIL$_NOSIGNAL,                 NULL,       NULL},
	{0,                0,                              NULL,       NULL}};

    item copy_out[]  = {
	{sizeof(newfile),  MAIL$_MESSAGE_FILE_CREATED,     &newfile,   NULL},
	{sizeof(newfold),  MAIL$_MESSAGE_FOLDER_CREATED,   &newfold,   NULL},
	{0,                0,                              NULL,       NULL}};

    DBGPRINT(21,"__copy_sequential_mail_messages: entered, seq file='%s', mailfile='%s', folder='%s'\n", seqfile, mailfile, folder, 0);

    DBGPRINT(25,"__copy_sequential_mail_messages: Begin mail context",0,0,0,0);
    if (file_ctx = __create_mail_context(seqfile))
      if (msg_context = __create_msg_context(&file_ctx))
        {
          DBGPRINT(25,"__copy_sequential_mail_messages: Begin message copy",0,0,0,0);
          for (id = 1; id <= msg_count; id++)
            {
              DBGPRINT(25,"__copy_sequential_mail_messages: copying message %d", id,0,0,0);
              status = mail$message_copy(&msg_context, copy_in, copy_out);
              if ( !chkstat("mail$message_copy", 0, status) )
                {
                  DBGPRINT(21,"__copy_sequential_mail_messages: exit, copy failed, status=0",0,0,0,0);
                  __vms_mail_close(&file_ctx,&msg_context);
                  return NIL;
                }

              if (newfile)
                {
                  DBGPRINT(25,"__copy_sequential_mail_messages: created mail file",0,0,0,0);
                }

              if (newfold)
                {
                  DBGPRINT(25,"__copy_sequential_mail_messages: created folder",0,0,0,0);
                }
            }
        }

    __vms_mail_close(&file_ctx, &msg_context);

    DBGPRINT(21,"__copy_sequential_mail_messages: exit, status=T",0,0,0,0);
    return T;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
   Purpose:
	Append a message stored in a STRING structure to a nominated mailbox

   Inputs:
	stream		A mail stream
	mailbox		The mailbox to which is to be appended
	flags		Pointer to list of flags to be set
	date		Pointer to date of message (can be NULL!)
	message		The STRING structure holding the message

   Outputs:
	NONE

   Return Value:
	T		If message appended successfully
	NIL		If not

   Comments:
	'append' in VMS MAIL terms actually means copying the message into
	a particular folder. Its position will be based on the date.

	Appending a message to an arbitrary folder in VMS MAIL is non-trivial.
	Essentially, what we do is to write a temporary file in the form of a
	VMS sequential mail file, then use the mail copy function to get it
	into the required folder.

	The format of a sequential mail file has to be fairly precise but is
        not obviously documented. I've worked it out by trial and error to be
	this (though of course it may change in the future!):

	    From:<tab>[<from-address><spaces>][<personal-name><spaces>]<date/time>
	    To:<tab>[<to-address>]
	    CC:<tab>[<cc-address>]
	    Subj:<tab>[<subject text>]
	    <blank line>
	    <text body>

	<..> items are data to be filled in
	[..] items are optional

	The data in the string structure is simply the text as received
	(normally the concatenation of the RFC822 format headers and the text
	body) so the headers must be parsed to extract the above information.

        The message text may contain '\r' characters. They are stripped out
	otherwise they confuse mail's header parsing. The '\n' is left alone
	though.

   Limitations:
	Currently, I can see no way to action the flags setting as, once the
        message has been copied into the mail folder, you have no means to
        locate it in order to use the message_modify function. Nor can you
	supply the flags setting at the time of the copy. This may be a 
	permanent restriction.

   Author:
	Andy Harper, Kings College London, January 1997
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
long vms_mail_append (MAILSTREAM *stream, char *mailbox, char *flags, char *date, STRING *message)
  {
    static char *months[] = {
	"JAN","FEB","MAR","APR","MAY","JUN","JUL","AUG","SEP","OCT","NOV","DEC"
	};
    static char *filepat[] = {
	"SYS$SCRATCH:SEQMAIL%08x.TMP",
	"SYS$LOGIN:SEQMAIL%08x.TMP",
	"SYS$DISK:[]SEQMAIL%08x.TMP"
			  };
    static int        maxfiles     = sizeof(filepat) / sizeof(filepat[0]);
    short             f            = vms_mail_getflags(stream,flags);
    long int          size         = SIZE(message);

    FILE              *fp;		/* for sequential mail file */
    ENVELOPE          *env;
    ADDRESS           *ap;
    MESSAGECACHE      elt;
    char              filename[MAILFILESIZE],mailfile[MAILFILESIZE],folder[FOLDERSIZE];
    char              address[256],datebuffer[256];
    char              c,*hdr,*ptr,*line,*tmp;
    int               hdrsize,inbox,i;

    DBGPRINT(21,"vms_mail_append: entered, mailbox='%s', flags='%s', size='%d'", mailbox, flags, size, 0);

					/* If no date, use today's date */
    if (!date)
      { rfc822_date(datebuffer); date = datebuffer; }

					/* Get fields of date */
    DBGPRINT(21,"vms_mail_append: date='%s'", date, 0, 0, 0);
    if (!mail_parse_date(&elt,date))
     {
       DBGPRINT(25,"vms_mail_append: date parse of '%s' failed",date,0,0,0);
       mm_log("unparseable date field",ERROR);
       return NIL;
     }
					/* Create VMS format date */
    sprintf(datebuffer,"%d-%s-%04d %2d:%02d:%02d",
      elt.day,months[elt.month-1],elt.year+BASEYEAR,elt.hours,elt.minutes,elt.seconds);


    DBGPRINT(25,"vms_mail_append: VMS format date ='%s'", datebuffer, 0, 0, 0);

					/* parse mailbox name */
    if (!__folder_exists(mailbox,NULL,1,mailfile,folder,&inbox))
      {
        char buf[256];
        sprintf(buf,"no such mailbox '%s'",mailbox);
        mm_log(buf,ERROR);
        return NIL;
      }

					/* Init to scan message */
    line=ptr=hdr = (char *) fs_get(size+1);


					/* Accumulate header until blank line */
    while (size--)
      if ((c=SNX(message)) != '\r')	/* Ignore \r characters */
        if ((*ptr++ = c) == '\n')
          {
            DBGPRINT(25,"vms_mail_append: line '%.*s'", ptr-line, line, 0, 0);
            if (ptr-line <= 1) break;	/* If blank line, stop here */
            line=ptr;			/* .. else note start of new line */
           }

					/* Parse the RFC822 headers */
    DBGPRINT(25,"vms_mail_append: parsing headers", 0, 0, 0, 0);
    hdrsize = ptr-hdr;
    tmp = (char *) fs_get(hdrsize + 1);
    rfc822_parse_msg(&env, NULL, hdr, hdrsize, NULL, mylocalhost(), tmp);
    fs_give((void **) &tmp);

					/* Try to open a scratch file */
    DBGPRINT(25,"vms_mail_append: opening a scratch file (%d attempts)", maxfiles, 0, 0, 0);
    for (i=0; i<maxfiles; i++)
      {
        sprintf(filename, filepat[i], getpid());
        DBGPRINT(25,"vms_mail_append: trying '%s'", filename, 0, 0, 0);
        if ((fp=fopen(filename, "w", "rfm=var","rat=cr")) != NULL) break;
        DBGPRINT(25,"vms_mail_append: failed to open '%s'", filename, 0, 0, 0);
      }

    if (fp == NULL)
      { DBGPRINT(21,"vms_mail_append: exit, status=0 : cant write seq mail file",0,0,0,0);
        mm_log("cannot write sequential mail file",ERROR);
        fs_give((void **)&hdr);
        return NIL;
      }

					/* Determine transport prefix */
     DBGPRINT(25,"vms_mail_append: Looking for transport prefix",0,0,0,0);
     if ((tmp=__transport_name()) != NULL)
       {
         sprintf(address,"%s%%%%\"%%s@%%s\"%%s",tmp);
         DBGPRINT(25,"vms_mail_append: prefix='%s', address format='%s'",tmp,address,0,0);
       }
     else
       {
         strcpy(address,"%s@%s%s");
         DBGPRINT(25,"vms_mail_append: no prefix, address format '%s'", address,0,0,0);
       }

					/* Write the VMS headers */
     DBGPRINT(25,"vms_mail_append: writing VMS headers to %s",filename,0,0,0);
     fputc('\f',fp);
     fputc('\n',fp);

     DBGPRINT(25,"vms_mail_append: processing from: header",0,0,0,0);
     fprintf(fp,"From:\t");
     for (ap=env->from; ap ; ap=ap->next)
      { DBGPRINT(25,"vms_mail_append: FROM: <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fprintf(fp," %s\n", datebuffer);

     DBGPRINT(25,"vms_mail_append: processing to: header",0,0,0,0);
     fprintf(fp,"To:\t");
     for (ap=env->to; ap ; ap=ap->next)
      { DBGPRINT(25,"vms_mail_append: TO:   <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fputc('\n',fp);

     DBGPRINT(25,"vms_mail_append: processing cc: header",0,0,0,0);
     fprintf(fp,"CC:\t");
     for (ap=env->cc; ap ; ap=ap->next)
      { DBGPRINT(25,"vms_mail_append: CC:   <%s@%s>", ap->mailbox,ap->host,0,0);
        fprintf(fp, address, ap->mailbox, ap->host, ap->next ?  "," : "");}
     fputc('\n',fp);

     DBGPRINT(25,"vms_mail_append: processing subj: header",0,0,0,0);
     fprintf(fp,"Subj:\t%s\n", env->subject ? env->subject : "");
     fputc('\n',fp);

					/* Free up the envelope */
     DBGPRINT(25,"vms_mail_append: free envelope structure",0,0,0,0);
     mail_free_envelope(&env);

					/* Write the RFC822 headers */
     DBGPRINT(25,"vms_mail_append: writing RFC822 headers (size='%d')\n%.*s",hdrsize,hdrsize,hdr,0);
     for (ptr=hdr; hdrsize-- > 0; ptr++)
       fputc(*ptr,fp);
     fs_give((void **)&hdr);

					/* Copy and write the body */
     DBGPRINT(25,"vms_mail_append: writing message text (size=%d)",size,0,0,0);
     while (size-- > 0)
       {c=SNX(message); if (c != '\r') fputc(c, fp);}


     DBGPRINT(25,"vms_mail_append: closing sequential mail file",0,0,0,0);
     fclose(fp);

					/* Add message to folder */
     DBGPRINT(25,"vms_mail_append: copying message into '%s'",mailbox,0,0,0);
     if (!__copy_sequential_mail_messages(filename, 1, mailfile, folder))
       { DBGPRINT(25,"vms_mail_append: append to folder failed",0,0,0,0);
         mm_log("append to folder failed",ERROR);
         remove(filename);
         return NIL;
       }

     DBGPRINT(25,"vms_mail_append: deleting temporary file '%s'",filename,0,0,0);
     remove(filename);

					/* Return appropriate status */
    DBGPRINT(21,"vms_mail_append: exit,status=1",0,0,0,0);
    return T;
}

/* VMS/MAIL garbage collect stream
 * Accepts: Mail stream
 *	    garbage collection flags
 */

void vms_mail_gc (MAILSTREAM *stream, long gcflags)
{
  /* nothing here for now */
}

/* Search support routines
 * Accepts: MAIL stream
 *	    message number
 *	    pointer to additional data
 * Returns: T if search matches, else NIL
 */


char vms_mail_search_all (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return T;			/* ALL always succeeds */
}


char vms_mail_search_answered (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->answered ? T : NIL;
}


char vms_mail_search_deleted (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->deleted ? T : NIL;
}


char vms_mail_search_flagged (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->flagged ? T : NIL;
}


char vms_mail_search_keyword (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->user_flags & n ? T : NIL;
}


char vms_mail_search_new (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (elt->recent && !elt->seen) ? T : NIL;
}



char vms_mail_search_old (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->recent ? NIL : T;
}


char vms_mail_search_recent (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->recent ? T : NIL;
}


char vms_mail_search_seen (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->seen ? T : NIL;
}


char vms_mail_search_unanswered (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->answered ? NIL : T;
}


char vms_mail_search_undeleted (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->deleted ? NIL : T;
}


char vms_mail_search_unflagged (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->flagged ? NIL : T;
}


char vms_mail_search_unkeyword (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->user_flags & n ? NIL : T;
}


char vms_mail_search_unseen (MAILSTREAM *stream, long msgno, char *d, long n)
{
  return mail_elt (stream,msgno)->seen ? NIL : T;
}



char vms_mail_search_before (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) < n);
}


char vms_mail_search_on (MAILSTREAM *stream, long msgno, char *d, long n)
{
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) (((elt->year << 9) + (elt->month << 5) + elt->day) == n);
}


char vms_mail_search_since (MAILSTREAM *stream, long msgno, char *d, long n)
{
				/* everybody interprets "since" as .GE. */
  MESSAGECACHE *elt = mail_elt (stream,msgno);
  return (char) ((long)((elt->year << 9) + (elt->month << 5) + elt->day) >= n);
}

char vms_mail_search_body (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char ret;
  char *s = vms_mail_fetchtext (stream,msgno);	/* get the text in LOCAL->buf */

  ret = search (s,strlen(s),d,n);/* do the search */
  return ret;			/* return search value */
}

char vms_mail_search_subject (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char *s = vms_mail_fetchstructure (stream,msgno,NIL)->subject;
  return s ? search (s,(long) strlen (s),d,n) : NIL;
}


char vms_mail_search_text (MAILSTREAM *stream, long msgno, char *d, long n)
{
  char ret;
  char *s = vms_mail_fetchheader (stream,msgno);	/* Get the header in LOCAL->buf */

  ret = search (s,strlen(s),d,n);  /* First, search at the header */
  if(!ret)	/* Not found at header, search at the body */
	  ret = vms_mail_search_body (stream,msgno,d,n);
  return ret;			/* return search value */
}



char vms_mail_search_bcc (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->bcc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_cc (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->cc;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_from (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->from;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}


char vms_mail_search_to (MAILSTREAM *stream, long msgno, char *d, long n)
{
  ADDRESS *a = vms_mail_fetchstructure (stream,msgno,NIL)->to;
  LOCAL->buf[0] = '\0';		/* initially empty string */
				/* get text for address */
  rfc822_write_address (LOCAL->buf,a);
  return search (LOCAL->buf,(long) strlen (LOCAL->buf),d,n);
}

/* Search parsers */


/* Parse a date
 * Accepts: function to return
 *	    pointer to date integer to return
 * Returns: function to return
 */

search_t vms_mail_search_date (search_t f, long *n)
{
  long i;
  char *s;
  MESSAGECACHE elt;
				/* parse the date and return fn if OK */
  return (vms_mail_search_string (f,&s,&i) && mail_parse_date (&elt,s) &&
	  (*n = (elt.year << 9) + (elt.month << 5) + elt.day)) ? f : NIL;
}



/* Parse a flag
 * Accepts: function to return
 *	    pointer to keyword integer to return
 *	    MAIL stream
 * Returns: function to return
 */

search_t vms_mail_search_flag (search_t f, long *n,MAILSTREAM *stream)
{
  short i;
  char *s,*t;
  if (t = strtok (NIL," ")) {	/* get a keyword */
    ucase (t);			/* get uppercase form of flag */
    for (i = 0; i < NUSERFLAGS && (s = stream->user_flags[i]); ++i)
      if (!strcmp (t,ucase (strcpy (LOCAL->buf,s))) && (*n = 1 << i)) return f;
  }
  return NIL;			/* couldn't find keyword */
}

/* Parse a string
 * Accepts: function to return
 *	    pointer to string to return
 *	    pointer to string length to return
 * Returns: function to return
 */


search_t vms_mail_search_string (search_t f,char **d, long *n)
{
  char *end = " ";
  char *c = strtok (NIL,"");	/* remainder of criteria */
  if (!c) return NIL;		/* missing argument */
  switch (*c) {			/* see what the argument is */
  case '{':			/* literal string */
    *n = strtol (c+1,d,10);	/* get its length */
    if ((*(*d)++ == '}') && (*(*d)++ == '\015') && (*(*d)++ == '\012') &&
	(!(*(c = *d + *n)) || (*c == ' '))) {
      char e = *--c;
      *c = DELIM;		/* make sure not a space */
      strtok (c," ");		/* reset the strtok mechanism */
      *c = e;			/* put character back */
      break;
    }
  case '\0':			/* catch bogons */
  case ' ':
    return NIL;
  case '"':			/* quoted string */
    if (strchr (c+1,'"')) end = "\"";
    else return NIL;
  default:			/* atomic string */
    if (*d = strtok (c,end)) *n = strlen (*d);
    else return NIL;
    break;
  }
  return f;
}

static unsigned int SendContext = 0;


/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Purpose:
	Check error status, cancel the current send and return pointer to
	suitable error message.

     Author:
	Andy Harper, Kings College London, Oct 1996
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
*/
static char *chksndstat(char *message, int status)
  {
     static char errmess[MAXRECORDSIZE];

     DBGPRINT(23,"%s: exit, status=0x%08x", message, status, 0, 0);

     if ((status & 0x1) == 0)
       {
         sprintf(errmess,"%s: failed with status 0x%08x", message, status);
         if (SendContext != 0) mail$send_end(&SendContext, NullItemList, NullItemList);
         return errmess;
       }

     return NULL;
  }

/*
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 | Get the message's text filename, the addresses and send it using the
 | selected mail protocol.
 | Returns NULL in case of success, or a pointer to the error message.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 */
static int MessageSentCount;	/* Set by the error/success routine */

char *vms_mail_send(char *filename, ADDRESS *from, ADDRESS *to, ADDRESS *cc, char *subject)
  {
    static char line[MAXRECORDSIZE];
    static short int typesu = MAIL$_TO, typecc = MAIL$_CC;
    ADDRESS *ap;
    FILE *ifd;
    int status, AddressesCount, success_send_result();
    char *err,*p, MailProtocolUsed[256];

    item ToItem[] = {
	{ 0,              MAIL$_SEND_USERNAME,      line,               NULL},
	{ sizeof(typesu), MAIL$_SEND_USERNAME_TYPE, &typesu,            NULL},
	{ 0, 0, NULL, NULL }};

    item CcItem[] = {
	{ 0,              MAIL$_SEND_USERNAME,      line,                NULL},
	{ sizeof(typecc), MAIL$_SEND_USERNAME_TYPE, &typecc,             NULL},
	{ 0, 0, NULL, NULL }};

    item FromItem[] = {
	{ 0,              MAIL$_SEND_FROM_LINE,     line,                NULL},
	{ 0, 0, NULL, NULL }};

    item SubjectItem[] = {
	{ 0,              MAIL$_SEND_SUBJECT,       line,                NULL},
	{ 0, 0, NULL, NULL }};

    item RecordItem[] = {
	{ 0,              MAIL$_SEND_RECORD,        line,                NULL},
	{ 0, 0, NULL, NULL }};

    item FileItem[] = {
	{ 0,              MAIL$_SEND_FILENAME,      line,                NULL},
	{ 0, 0, NULL, NULL }};

    item SendItem[] = {
	{ sizeof (int *), MAIL$_SEND_SUCCESS_ENTRY, success_send_result, NULL},
	{ 0,              MAIL$_NOSIGNAL,           NULL,                NULL},
	{ 0, 0, NULL, NULL }};

/* Get the foreign protocol to use */
    if((p = __transport_name()) != NULL)
      strcpy(MailProtocolUsed, p);
    else
      return "No foreign mail transport protocol defined";

/* Create mail context */
    SendContext = 0;
    status = mail$send_begin(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_begin", status)) != NULL) return err;

/* Create the sender's address */
/* Use the system default FROM: address (to avoid privileges)
    sprintf(line, "%s <%s@%s>", (from->personal ? from->personal : ""),
	from->mailbox, from->host);
    FromItem[0].length = strlen(line);
    status = mail$send_add_attribute(&SendContext, FromItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
*/

/* Now the subject */
    if(subject) {
	strcpy(line, subject);
	SubjectItem[0].length = strlen(line);
	status = mail$send_add_attribute(&SendContext, SubjectItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_attribute", status)) != NULL) return err;
    }

/* now add the TO addresses */
    AddressesCount = 0;
    ap = to;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	ToItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, ToItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Same for CC */
    ap = cc;
    while(ap != NULL) {
	sprintf(line, "%s%%\"%s@%s\"", MailProtocolUsed, ap->mailbox, ap->host);
	CcItem[0].length = strlen(line);
	status = mail$send_add_address(&SendContext, CcItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_address", status)) != NULL) return err;
	ap = ap->next;
	AddressesCount++;
    }

/* Open the message's file */
    if((ifd = fopen(filename, "r")) == NULL) {
	mail$send_end(&SendContext, NullItemList, NullItemList);
	sprintf(line, "Can't open '%s'", filename);
	return line;
    }

/* Now copy it. VMS/MAIL can accept up to MAXRECORDSIZE characters long lines */

/* --- Method 1: Add each record, one at a time
    while(fgets(line, (sizeof(line) < MAXRECORDSIZE ? sizeof(line) - 1 : MAXRECORDSIZE), ifd) != NULL) {
	char *p = strchr(line, '\n');
	if(p) *p = '\0';
	RecordItem[0].length = strlen(line);
	status = mail$send_add_bodypart(&SendContext, RecordItem, NullItemList);
       	if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;
    }
*/
    fclose(ifd);

/* --- Method 2: Just add the filename */
    strcpy(line,filename);
    FileItem[0].length = strlen(line);
    status = mail$send_add_bodypart(&SendContext, FileItem, NullItemList);
    if ( (err = chksndstat("mail$send_add_bodypart", status)) != NULL) return err;



/* OK, done - actually send the message. This routine will call via AST a routine
   that will count how many were sent ok.
*/
    MessageSentCount = 0;	/* None set yet */
    status = mail$send_message(&SendContext, SendItem, NullItemList);
    if ( (err = chksndstat("mail$send_message", status)) != NULL) return err;

    status = mail$send_end(&SendContext, NullItemList, NullItemList);
    if ( (err = chksndstat("mail$send_end", status)) != NULL) return err;

    if(MessageSentCount < AddressesCount)
      {
        switch (MessageSentCount)
          {
            case 0: sprintf(line,"error - no messages sent"); break;
            case 1: sprintf(line,"error - 1 message sent, %d failed",AddressesCount-1); break;
            default:sprintf(line,"error - %d messages sent, %d failed",MessageSentCount, AddressesCount); break;
          }
        return line;
      }
    return NULL;	/* Success */
}




/*
 | this routine is called only if the message was sent ok. We count the number
 | of messages sent ok.
 */
int success_send_result(item *recipient, int *sigarray, int userdata)
  {
    MessageSentCount++;	/* OK */
  }
